/* ircd.c -- start-up code and some core functions;
 * Copyright (C) 2001-2002, ForestNet IRC Network
 * This is a free software; see file `license' for details.
 */

#ifndef lint
static char cvsid[] = "$Id: ircd.c,v 1.43 2002/07/14 12:18:25 hex Exp $";
#endif

static const char *usage_text =
#ifndef _WIN32
"Startup options, only available to superuser (faults are fatal):\n"
"  -u name|id   change effective user id to this\n"
"  -g id        change effective group id to this\n"
"  -r path      change filesystem root to this\n"
"\n"
#endif
"Configuration overrides:\n"
"  -f file      use this instead of ./ircd.conf\n"
"  -h name      override server name\n"
"  -p port      listen to this port number for server connections\n"
"\n"
#ifdef DEBUGMODE
"Debugging:\n"
"  -x level     set debug.log verbosity\n"
"\n"
#endif
#ifndef _WIN32
"Misc options:\n"
"  -t           do not detach from console\n"
"\n"
#endif
;

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "options.h"

#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>

#ifndef _WIN32
# include <sys/file.h>
# include <pwd.h>
# include <sys/time.h>
#else
# include <io.h>
# include <direct.h>
#endif

#ifdef HPUX
# define _KERNEL /* HPUX has the world's worst headers... */
#endif

#ifndef _WIN32
# include <sys/resource.h>
#endif

#ifdef HPUX
# undef _KERNEL
#endif

#include <errno.h>
#include "h.h"

#if defined(_WIN32) && defined(DEBUGMODE)
# include <stdio.h>
#endif

#include <string.h>
#include "unicode.h"

#include "channel.h" /* for MODEPARAMS */

int R_do_dns, R_fin_dns, R_fin_dnsc, R_fail_dns, R_do_id, R_fin_id, R_fail_id;
char REPORT_DO_DNS[128], REPORT_FIN_DNS[128], REPORT_FIN_DNSC[128],
	REPORT_FAIL_DNS[128], REPORT_DO_ID[128], REPORT_FIN_ID[128],
	REPORT_FAIL_ID[128];

char buf[BUFSIZE];

aClient me; /* That's me */
aClient *client = &me; /* Pointer to beginning of Client list */

void server_reboot(char *);
void restart (char *);
static void open_debugfile(), setup_signals();
void check_lusers(void);

char *configfile = "ircd.conf"; /* default configuration file name */

/* default options, overriden in ircd.conf */
char *network_name = NULL;
char *kline_address = NULL;
char *contact_url = NULL;
char *services_host = NULL;
char *server_mask = NULL;
char *ns_nick = NULL, *cs_nick = NULL, *ms_nick = NULL, *os_nick = NULL;
char *codepage_path = NULL;
char *opers_log = NULL;
int def_umode = 0;
int nickchange_delay = 0;
int show_connect_info = 0;
int force_motd = 0;

/* default debug file name */
#ifdef DEBUGMODE
char *debugfn = "debug.log";
#else
# if VMS
#  define LOGFILE "NLA0:"
# else
#  define LOGFILE "/dev/null"
# endif
#endif

int portnum = -1; // Server port number, listening this

int debuglevel = -1;                // Server debug level
int bootopt = 0;                    // Server boot option flags
char *debugmode = "";               // -"-    -"-   -"-
char *sbrk0;                        // initial sbrk(0)
static int dorehash = 0;
char **myargv;

time_t nextconnect = 1;        // time for next try_connections call
time_t nextping = 1;           // same as above for check_pings()
time_t nextdnscheck = 0;       // next time to poll dns to force timeouts
time_t nextexpire = 1;         // next expire run on the dns cache
time_t lastlucheck = 0;
int    lu_noninv, lu_inv, lu_serv, lu_oper, lu_unknown, lu_channel,
	   lu_lu, lu_lserv, lu_clu, lu_mlu, lu_cglobalu, lu_mglobalu;

int oper_closed = 0; /* server unusable */

#ifdef CLONE_CHECK
		aClone *Clones = NULL;
		char clonekillhost[100];
#endif

time_t NOW;
#if    defined(PROFIL) && !defined(_WIN32)
extern etext();

VOIDSIG s_monitor()
{
	static int mon = 0;
#ifdef POSIX_SIGNALS
	struct sigaction act;
#endif

	(void)moncontrol(mon);
	mon = 1 - mon;
#ifdef POSIX_SIGNALS
	act.sa_handler = s_rehash;
	act.sa_flags = 0;
	(void)sigemptyset(&act.sa_mask);
	(void)sigaddset(&act.sa_mask, SIGUSR1);
	(void)sigaction(SIGUSR1, &act, NULL);
#else
	(void)signal(SIGUSR1, s_monitor);
#endif
}
#endif

VOIDSIG s_die()
{
#ifdef USE_SYSLOG
	(void)syslog(LOG_CRIT, "Server Killed By SIGTERM");
#endif
	flush_connections(me.fd);
	/* Borg88
	Probably here should be WSACleanup() but adding this, makes that server
	crash sometimes (for me happend only once) after WM_ENDSESSION and probably
	by WM_QUIT and /die too
	*/
#ifdef _WIN32
	Sleep(1000);
#else
	sleep(1);
#endif
	exit(0);
}

#ifndef _WIN32
static VOIDSIG s_rehash()
#else
VOIDSIG s_rehash()
#endif
{
#ifdef POSIX_SIGNALS
	   struct  sigaction act;
#endif
	   dorehash = 1;
#ifdef POSIX_SIGNALS
	   act.sa_handler = s_rehash;
	   act.sa_flags = 0;
	   (void)sigemptyset(&act.sa_mask);
	   (void)sigaddset(&act.sa_mask, SIGHUP);
	   (void)sigaction(SIGHUP, &act, NULL);
#else
# ifndef _WIN32
	   (void)signal(SIGHUP, s_rehash); // sysV -argv
# endif
#endif
}

void restart(char *mesg)
{
#ifdef USE_SYSLOG
	syslog(LOG_WARNING, "Restarting Server because: %s",mesg);
#endif
	server_reboot(mesg);
}

VOIDSIG s_restart()
{
	static int restarting = 0;

#ifdef USE_SYSLOG
	(void)syslog(LOG_WARNING, "Server Restarting on SIGINT");
#endif
	if (restarting == 0) {
		/* Send (or attempt to) a dying scream to oper if present */
		restarting = 1;
		server_reboot("SIGINT");
	}
}

void server_reboot(char *mesg)
{
	int i;

	sendto_ops("Aieeeee!!!  Restarting server... %s", mesg);
	Debug((DEBUG_NOTICE,"Restarting server... %s", mesg));
	flush_connections(me.fd);

	/* fd 0 must be 'preserved' if either the -d or -i options have
	 * been passed to us before restarting. */
#ifdef USE_SYSLOG
	closelog();
#endif
#ifndef _WIN32
	for (i = 3; i < MAXCONNECTIONS; i++)
		close(i);
	if (!(bootopt & (BOOT_TTY|BOOT_DEBUG)))
		close(2);
	close(1);
	if ((bootopt & BOOT_CONSOLE) || isatty(0))
		close(0);
	if (!(bootopt & (BOOT_INETD|BOOT_OPER)))
		execv("ircd", myargv);
#else
	for (i = 0; i < highest_fd; i++)
		if ( closesocket(i) == -1 ) close(i);

	execv(myargv[0], myargv);
#endif
#ifdef USE_SYSLOG
	/* Have to reopen since it has been closed above */

	openlog(myargv[0], LOG_PID|LOG_NDELAY, LOG_FACILITY);
	syslog(LOG_CRIT, "execv(%s,%s) failed: %m\n", "ircd", myargv[0]);
	closelog();
#endif

#ifndef _WIN32
	Debug((DEBUG_FATAL, "Couldn't restart server: %s", strerror(errno)));
#else
	Debug((DEBUG_FATAL, "Couldn't restart server: %s",
		strerror(GetLastError())));
#endif

	exit(-1);
}


/* Scan through configuration and try new connections.  Returns the calendar
 * time when the next call to this function should be made latest. (No harm
 * done if this is called earlier or later...)
 * Borg88: Rewrited. No more stats c cycle.
 */
static time_t try_connections(time_t currenttime)
{
	aConfItem *aconf, *cconf, *con_conf;
	aClass *cltmp;
	aClient *cptr;
	int connecting, confrq, found, con_class = 0;
	time_t next = 0;

	connecting = FALSE;
	found = 0;
	/*
	Debug((DEBUG_NOTICE, "Connection check at: %s", myctime(currenttime)));
	*/
	for (aconf = conf; aconf; aconf = aconf->next) {
		/* Also when already connecting! (update holdtimes) --SRB */
		if (!(aconf->status & CONF_CONNECT_SERVER) || aconf->port <= 0)
			continue;
		if (aconf->flags & FLAG_TRIED) {
			found = 1;
			continue;
		}
		cltmp = Class(aconf);
		/*
		** Skip this entry if the use of it is still on hold until
		** future. Otherwise handle this entry (and set it on hold
		** until next time). Will reset only hold times, if already
		** made one successfull connection... [this algorithm is
		** a bit fuzzy... -- msa >;) ]
		*/

		if ((aconf->hold > currenttime)) {
			if ((next > aconf->hold) || (next == 0))
				next = aconf->hold;
			continue;
		}

		confrq = get_con_freq(cltmp);
		aconf->hold = currenttime + confrq;

		/* Found a CONNECT config with port specified, scan clients
		 * and see if this server is already connected? */
		cptr = find_name(aconf->name, (aClient *)NULL);

		if (!connecting && !cptr && Links(cltmp) < MaxLinks(cltmp)) {
			/* Check connect rules to see if we're allowed to try */
			for (cconf = conf; cconf; cconf = cconf->next) {
				if ((cconf->status & CONF_CRULE)
				&& (match(cconf->host, aconf->name) == 0)) {
					if (crule_eval(cconf->passwd))
					break;
				}
			}
			
			if (!cconf) {
				con_class = Class(cltmp);
				con_conf = aconf;
				/* We connect only one at time... */
				connecting = TRUE;
			}
		}

		if (cptr || Links(cltmp)>=MaxLinks(cltmp))
			aconf->flags|=FLAG_TRIED;

		if((next > aconf->hold) || (next == 0))
			next = aconf->hold;
	}

	if (connecting) {
		if (connect_server(con_conf,
		(aClient *)NULL,(struct hostent *)NULL) == 0) {
			sendto_ops("Connection to %s[%s] activated.",
				con_conf->name, con_conf->host);
			con_conf->flags |= FLAG_TRIED;
			next = currenttime + 100;
		}
	}

	/* Debug((DEBUG_NOTICE,"Next connection check : %s", myctime(next))); */
	if (found && !next) {
		for (aconf = conf; aconf; aconf = aconf->next) {
			if (aconf->status & CONF_CONNECT_SERVER)
				aconf->flags &= ~FLAG_TRIED;
		}
		return 1;
	}

	return next;
}


/* Now find_kill is only called when a kline-related command is used:
   AKILL/RAKILL/KLINE/UNKLINE/REHASH.  Very significant CPU usage decrease.
   I made changes to evm_lusers
ery check_pings call to add new parameter.
   -- Barubary */
time_t check_pings(time_t currenttime, int check_kills)
{
	aClient *cptr;
	int killflag, ping = 0, i, rflag = 0;
	time_t oldest = 0, timeout;

#ifdef TIMED_KLINES
	check_kills = 1;
#endif

	/* reversed scan to handle remaps */
	for (i = highest_fd; i >= 0; --i) {
		/* adjust if remapped */
		if (i > highest_fd)
			i = highest_fd;

		if (!(cptr = local[i]) || IsMe(cptr) || IsLog(cptr))
			continue;

		/* Note: No need to notify opers here. It's already done when
		  "FLAGS_DEADSOCKET" is set. */
		if (cptr->flags & FLAGS_DEADSOCKET) {
			exit_client(cptr, cptr, &me, "Dead socket");
			continue;
		}

		killflag = (check_kills) ? (IsPerson(cptr) ? find_kill(cptr) : 0) : 0;

		if (check_kills && !killflag && IsPerson(cptr))
			if (find_zap(cptr))
				killflag = 1;

#ifdef R_LINES_OFTEN
			rflag = IsPerson(cptr) ? find_restrict(cptr) : 0;
#endif

			ping = IsRegistered(cptr) ? get_client_ping(cptr)
				: CONNECTTIMEOUT;
			Debug((DEBUG_DEBUG, "c(%s)=%d p %d k %d r %d a %d",
				cptr->name, cptr->status, ping, killflag, rflag,
				currenttime - cptr->lasttime));

			/* Ok, so goto's are ugly and can be avoided here but this code
			   is already indented enough so I think its justified. -avalon */
			if (!killflag && !rflag && IsRegistered(cptr) &&
				(ping >= currenttime - cptr->lasttime))
					goto ping_timeout;

			/* If the server hasnt talked to us in 2*ping seconds
			   and it has a ping time, then close its connection.
			   If the client is a user and a KILL line was found
			   to be active, close this connection too. */
			if (killflag || rflag || ((currenttime - cptr->lasttime) >= (2 * ping) && (cptr->flags & FLAGS_PINGSENT)) ||
			   (!IsRegistered(cptr) && (currenttime - cptr->firsttime) >= ping)) {
				if (!IsRegistered(cptr) && (DoingDNS(cptr) || DoingAuth(cptr))) {
					if (cptr->authfd >= 0) {
#ifndef _WIN32
						close(cptr->authfd);
#else
						closesocket(cptr->authfd);
#endif
						cptr->authfd = -1;
						cptr->count = 0;
						*cptr->buffer = '\0';
					}
					if (show_connect_info) {
						if (DoingDNS(cptr))
							write(cptr->fd, REPORT_FAIL_DNS, R_fail_dns);
						else
							write(cptr->fd, REPORT_FAIL_ID, R_fail_id);
					}
					Debug((DEBUG_NOTICE, "DNS/AUTH timeout %s", get_client_name(cptr,TRUE)));
					del_queries((char *)cptr);
					ClearAuth(cptr);
					ClearDNS(cptr);
					SetAccess(cptr);
					cptr->firsttime = currenttime;
					cptr->lasttime = currenttime;
					continue;
				}

				if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr)) {
					sendto_ops("No response from %s, closing link", get_client_name(cptr, FALSE));
					sendto_serv_butone(&me, ":%s GNOTICE :No response from %s, closing link", me.name, get_client_name(cptr, FALSE));
				}

				/* this is used for KILL lines with time restrictions on
				   them - send a messgae to the user being killed first. */
				if (killflag && IsPerson(cptr))
					sendto_ops("Kill line active for %s[%s@%s]", cptr->name, cptr->username, (*cptr->vhost) ? cptr->vhost : cptr->sockhost);

#if defined(R_LINES) && defined(R_LINES_OFTEN)
				if (IsPerson(cptr) && rflag)
					sendto_ops("Restricting %s, closing link.", get_client_name(cptr,FALSE));
#endif
				exit_client(cptr, cptr, &me, killflag ? gcomment : "Ping timeout");

				continue;
			}
			else if (IsRegistered(cptr) && (cptr->flags & FLAGS_PINGSENT) == 0) {
				/* if we havent PINGed the connection and we havent heard
				   from it in a while, PING it to make sure it is still
				   alive. */
				cptr->flags |= FLAGS_PINGSENT;

				/* not nice but does the job */
				cptr->lasttime = currenttime - ping;
				sendto_one(cptr, "PING :%s", me.name);
			}

ping_timeout:
		timeout = cptr->lasttime + ping;
		while (timeout <= currenttime)
			timeout += ping;
		if (timeout < oldest || !oldest)
			oldest = timeout;
	}
	if (!oldest || oldest < currenttime)
		oldest = currenttime + PINGFREQUENCY;

	/*
	Debug((DEBUG_NOTICE,"Next check_ping() call at: %s, %d %d %d",
		myctime(oldest), ping, oldest, currenttime));
	*/

	return (oldest);
}

/*
** bad_command
**     This is called when the commandline is not acceptable.
**     Give error message and exit without starting anything.
*/
static int bad_command()
{
#ifndef _WIN32
	printf("Usage: ircd [OPTION]...\n\n%s", usage_text);
#else
	MessageBox(NULL, "Usage: faerion [OPTION]...\n%s",
		usage_text, MB_OK);
#endif
	exit(-1);
}


#ifndef _WIN32
int main(int argc, char *argv[])
#else
int InitwIRCD(int argc, char *argv[])
#endif
{
#ifdef _WIN32
	WORD wVersionRequested = MAKEWORD(1, 1);
	WSADATA wsaData;
#else
	uid_t uid, euid, want_uid = 0;
	gid_t want_gid = 0;
	time_t delay = 0;
	char *want_root = NULL, fuckup[256];
	FILE *fdee;
#endif
	int portarg = 0;
#if defined(FORCE_CORE) && !defined(_WIN32)
	struct rlimit corelim;
#endif

#ifndef _WIN32
	sbrk0 = (char *)sbrk((size_t)0);
	uid = getuid();
	euid = geteuid();
# ifdef PROFIL
	(void)monstartup(0, etext);
	(void)moncontrol(1);
	(void)signal(SIGUSR1, s_monitor);
# endif
#endif

#ifndef _WIN32
	res_init();
#endif

	myargv = argv;
#ifndef _WIN32
	(void)umask(077); /* better safe than sorry --SRB */
#else
	/* WSAStartup(wVersionRequested, &wsaData); */
#endif

	bzero((char *)&me, sizeof(me));

#ifndef _WIN32
	setup_signals();
#endif

#if defined(FORCE_CORE) && !defined(_WIN32)
	corelim.rlim_cur = corelim.rlim_max = RLIM_INFINITY;
	if (setrlimit(RLIMIT_CORE, &corelim))
		printf("unlimit core size failed; errno = %d, no core will be "
			"dumped on an error.\n", errno);
#endif

#ifdef USE_CASETABLES
	/* Set up the case tables */
	setup_match();
#endif

	/*
	** All command line parameters have the syntax "-fstring"
	** or "-f string" (e.g. the space is optional). String may
	** be empty. Flag characters cannot be concatenated (like
	** "-fxyz"), it would conflict with the form "-fstring".
	*/
	while (--argc > 0 && (*++argv)[0] == '-') {
		char *p = argv[0]+1;
		int flag = *p++;

		if (flag == '\0' || *p == '\0')
			if (argc > 1 && argv[1][0] != '-') {
				p = *++argv;
				argc -= 1;
			}
			else
				p = "";

		switch (flag) {
#ifndef _WIN32
			case 'a':
				bootopt |= BOOT_AUTODIE;
				break;
			case 'c':
				bootopt |= BOOT_CONSOLE;
				break;
			case 'd' :
				(void)setuid((uid_t)uid);
				break;
			case 'q':
				bootopt |= BOOT_QUICK;
				break;
			case 'g':
				if (BadPtr(p) || !(want_gid = atoi(p))) {
					fprintf(stderr, "%s: must specify a "
						"real group id with -g\n",
						*myargv);
					exit(-1);
				}
				break;
			case 'u':
				if (BadPtr(p)) {
					fprintf(stderr, "%s: must specify a "
						"real user name/id with -u\n",
						*myargv);
					exit(-1);
				}
				if (!(want_uid = atoi(p))) {
					struct passwd *pw;
					if (pw = getpwnam(p)) {
						want_uid = pw->pw_uid;
						want_gid = pw->pw_gid;
					}
				}
				break;
			case 'r':
				if (BadPtr(p)) {
					fprintf(stderr, "%s: must specify a "
						"path to chroot with -r\n",
						*myargv);
					exit(-1);
				}
				want_root = p;
				break;
#else
			case 'd':
				break;
#endif
#ifndef _WIN32
			case 'o': /* Per user local daemon... */
				(void)setuid((uid_t)uid);
				bootopt |= BOOT_OPER;
				break;
#endif

			case 'f':
# ifndef _WIN32
				(void)setuid((uid_t)uid);
# endif
				configfile = p;
				break;

#ifndef _WIN32
			case 'i':
				bootopt |= BOOT_INETD|BOOT_AUTODIE;
				break;
#endif
			case 'h':
				strncpyzt(me.name, p, sizeof(me.name));
				break;

			case 'p':
				if ((portarg = atoi(p)) >= 1024)
					portnum = portarg;
				else {
					fprintf(stderr, "%s: please specify a "
						"valid port number with -p "
						"(must be >= 1024).\n",
						*myargv);
					exit(-1);
				}
				break;
#ifndef _WIN32
			case 't':
				setuid((uid_t)uid);
				bootopt |= BOOT_TTY;
				break;

			case 'v':
				printf("%s\n", version);
#else
			case 'v':
				MessageBox(NULL, version, "faerion version",
					MB_ICONINFORMATION);
#endif
				exit(0);

			case 'x':
#ifdef DEBUGMODE
# ifndef _WIN32
				setuid((uid_t)uid);
# endif
				debuglevel = atoi(p);
				debugmode = *p ? p : "0";
				bootopt |= BOOT_DEBUG;
				break;
#else /* DEBUGMODE */
# ifndef _WIN32
				fprintf(stderr, "%s: DEBUGMODE must be defined "
					"for -x, recompile?\n", myargv[0]);
# else
				MessageBox(NULL, "You can't use the -x option "
					"in a release version of daemon, "
					"please obtain a debug version.",
					argv[0], MB_OK|MB_ICONHAND);
# endif
				exit(0);
#endif

			default:
				bad_command();
				break;
		}
	}

#ifndef _WIN32
	if (!uid) {
		if (!want_uid || !want_gid) {
			fprintf(stderr, "ERROR: specify a real user/group id "
				"with -u and -g, when running as superuser.\n");
			exit(-1);
		}

		if (BadPtr(want_root))
			want_root = ".";
		if (chroot(want_root)) {
			fprintf(stderr, "ERROR: failed to change filesystem "
				"root to %s.\n", want_root);
			exit(-1);
		}
		else chdir("/");

		if (want_gid && (setegid(want_gid) || (setgid(want_gid)))) {
			fprintf(stderr, "ERROR: failed to change group id to "
				"%u.\n", want_gid);
			exit(-1);
		}

		if (want_uid && (seteuid(want_uid) || setuid(want_uid))) {
			fprintf(stderr, "ERROR: failed to change user id to "
				"%u.\n", want_uid);
			exit(-1);
		}
	}
	else if (want_uid || want_gid || !BadPtr(want_root)) {
		fprintf(stderr, "%s: you must be a superuser to "
			"use -r, -g and -r\n", *myargv);
		exit(-1);
	}

	if ((debuglevel < 0) &&  (bootopt & BOOT_TTY)) {
		fprintf(stderr, "you specified -t without -x. use -x <n>\n");
		exit(-1);
	}
#endif

	if (argc > 0)
		return bad_command(); /* This should exit out */

	clear_client_hash_table();
	clear_channel_hash_table();
	clear_notify_hash_table();
	inittoken();
	initlists();
	initclass();
	initwhowas();
	initstats();
	open_debugfile();
	if (portnum < 0)
		portnum = PORTNUM;
	me.port = portnum;
	(void)init_sys();
	me.flags = FLAGS_LISTEN;

#ifndef _WIN32
	if (bootopt & BOOT_INETD) {
		me.fd = 0;
		local[0] = &me;
		me.flags = FLAGS_LISTEN;
	}
	else
#endif
		me.fd = -1;

#ifdef USE_SYSLOG
	openlog(myargv[0], LOG_PID|LOG_NDELAY, LOG_FACILITY);
#endif

	if (initconf(bootopt) == -1) {
		char msg[256];
		snprintf(msg, sizeof(msg)-1, "Couldn't open configuration "
			"file %s", configfile);

		Debug((DEBUG_FATAL, msg));

#ifndef _WIN32
		(void)printf(msg);
#else
		MessageBox(NULL, msg, APPNAME, MB_OK|MB_ICONHAND);
#endif
		exit(-1);
	}

	if (!(bootopt & BOOT_INETD)) {
		static char star[] = "*";
		aConfItem *aconf;

		if ((aconf = find_me()) && portarg <= 0 && aconf->port > 0)
			portnum = aconf->port;
		Debug((DEBUG_ERROR, "Port = %d", portnum));
		if (inetport(&me, aconf->passwd, portnum))
			exit(1);
	}
	else if (inetport(&me, "*", 0))
		exit(1);

	(void)setup_ping();
	(void)get_my_name(&me, me.sockhost, sizeof(me.sockhost)-1);
	if (me.name[0] == '\0')
		strncpyzt(me.name, me.sockhost, sizeof(me.name));
	me.hopcount = 0;
	me.authfd = -1;
	me.confs = NULL;
	me.next = NULL;
	me.user = NULL;
	me.from = &me;
	SetMe(&me);
	make_server(&me);
	(void)strcpy(me.serv->up, me.name);

	me.lasttime = me.since = me.firsttime = time(NULL);
	(void)add_to_client_hash_table(me.name, &me);

	sprintf(REPORT_DO_DNS, ":%s %s", me.name, BREPORT_DO_DNS);
	sprintf(REPORT_FIN_DNS, ":%s %s", me.name, BREPORT_FIN_DNS);
	sprintf(REPORT_FIN_DNSC, ":%s %s", me.name, BREPORT_FIN_DNSC);
	sprintf(REPORT_FAIL_DNS, ":%s %s", me.name, BREPORT_FAIL_DNS);
	sprintf(REPORT_DO_ID, ":%s %s", me.name, BREPORT_DO_ID);
	sprintf(REPORT_FIN_ID, ":%s %s", me.name, BREPORT_FIN_ID);
	sprintf(REPORT_FAIL_ID, ":%s %s", me.name, BREPORT_FAIL_ID);

	R_do_dns = strlen(REPORT_DO_DNS);
	R_fin_dns = strlen(REPORT_FIN_DNS);
	R_fin_dnsc = strlen(REPORT_FIN_DNSC);
	R_fail_dns = strlen(REPORT_FAIL_DNS);
	R_do_id = strlen(REPORT_DO_ID);
	R_fin_id = strlen(REPORT_FIN_ID);
	R_fail_id = strlen(REPORT_FAIL_ID);

	check_class();
	if (bootopt & BOOT_OPER) {
		aClient *tmp = add_connection(&me, 0);

		if (!tmp)
			exit(1);
		SetMaster(tmp);
	}
	else
		write_pidfile();

	Debug((DEBUG_NOTICE,"Server ready..."));
#ifdef USE_SYSLOG
	syslog(LOG_NOTICE, "Server Ready");
#endif

#ifdef _WIN32
	return 1;
}


#if 0
void TerminateDaemon(void)
{
	int i;
	aClient *acptr;

	sendto_ops("Server Terminating from a local console", me.name);

	/* remove clients first */
	for (i = highest_fd; i; i--)
		if ((acptr = local[i]) && IsPerson(acptr))
			exit_client(acptr, acptr, acptr, "Server Terminated");

	/* disconnect servers */
	for (i = highest_fd; i; i--)
		if ((acptr = local[i]) && IsServer(acptr))
			sendto_one(acptr, ":%s ERROR :Terminated by local console", me.name);

	s_die();
	return;
}
#endif


void SocketLoop(void *dummy)
{
	time_t delay = 0, now;

	while (1)
#else
	for (;;)
#endif
	{
		now = time(NULL);
		// Run through the hashes and check lusers every second
		if (lastlucheck < now) {
			check_lusers();
			lastlucheck=now;
		}

		// We only want to connect if a connection is due,
		// not every time through.  Note, if there are no
		// active C lines, this call to Tryconnections is
		// made once only; it will return 0. - avalon
		if (nextconnect && now >= nextconnect)
			nextconnect = try_connections(now);

		// DNS checks. One to timeout queries, one for cache expiries.
		if (now >= nextdnscheck)
			nextdnscheck = timeout_query_list(now);
		if (now >= nextexpire)
			nextexpire = expire_cache(now);

		// take the smaller of the two 'timed' event times as
		// the time of next event (stops us being late :) - avalon
		// WARNING - nextconnect can return 0!
		if (nextconnect)
			delay = MIN(nextping, nextconnect);
		else
			delay = nextping;
		delay = MIN(nextdnscheck, delay);
		delay = MIN(nextexpire, delay);
		delay -= now;

		// Adjust delay to something reasonable [ad hoc values]
		// (one might think something more clever here... --msa)
		// We don't really need to check that often and as long
		// as we don't delay too long, everything should be ok.
		// waiting too long can cause things to timeout...
		// i.e. PINGS -> a disconnection :(
		// - avalon
		if (delay < 1)
			delay = 1;
		else
			delay = MIN(delay, TIMESEC);
		read_message(delay);

		Debug((DEBUG_DEBUG ,"Got message(s)"));

		now = time(NULL);
		// ...perhaps should not do these loops every time,
		// but only if there is some chance of something
		// happening (but, note that conf->hold times may
		// be changed elsewhere--so precomputed next event
		// time might be too far away... (similarly with
		// ping times) --msa
		if (now >= nextping)
			nextping = check_pings(now, 0);

		if (dorehash) {
			rehash(&me, &me, 1);
			dorehash = 0;
		}

		// Flush output buffers on all connections now if they
		// have data in them (or at least try to flush)
		// -avalon
		flush_connections(me.fd);
	}
}

/*
 * open_debugfile
 *
 * If the -t option is not given on the command line when the server is started
 * Here we just open that file and make sure it is opened to fd 2 so that
 * any fprintf's to stderr also goto the logfile.  If the debuglevel is not
 * set from the command line by -x, use /dev/null as the dummy logfile as long
 * as DEBUGMODE has been defined, else dont waste the fd.
 */

static void open_debugfile()
{
#ifdef DEBUGMODE
# ifdef _WIN32
	extern FILE *debugfile;
# endif
	int fd;
	aClient *cptr;

	if (debuglevel >= 0) {
#ifdef _WIN32
		debugfile = fopen(debugfn, "w");
#endif

		cptr = make_client(NULL, NULL);
		cptr->fd = 2;
		SetLog(cptr);
		cptr->port = debuglevel;
		cptr->flags = 0;
		cptr->acpt = cptr;
		local[2] = cptr;
		strcpy(cptr->sockhost, me.sockhost);
# ifndef _WIN32
		printf("isatty = %d ttyname = %#x\n", isatty(2), (u_int)ttyname(2));
		if (!(bootopt & BOOT_TTY)) { // leave debugging output on fd 2
			truncate(debugfn, 0);
			if ((fd = open(debugfn, O_WRONLY | O_CREAT, 0600)) < 0)
				if ((fd = open("/dev/null", O_WRONLY)) < 0)
					exit(-1);
			if (fd != 2) {
				dup2(fd, 2);
				close(fd);
			}
			strncpyzt(cptr->name, debugfn, sizeof(cptr->name));
		}
		else if (isatty(2) && ttyname(2))
			strncpyzt(cptr->name, ttyname(2), sizeof(cptr->name));
		else
# endif // _WIN32
			strcpy(cptr->name, "FD2-Pipe");
		Debug((DEBUG_FATAL, "Debug: File <%s> Level: %d at %s", cptr->name, cptr->port, myctime(time(NULL))));
	}
	else
		local[2] = NULL;
#endif

	return;
}

#ifndef _WIN32
static void setup_signals()
{
#ifdef POSIX_SIGNALS
	struct  sigaction act;

	act.sa_handler = SIG_IGN;
	act.sa_flags = 0;
	(void)sigemptyset(&act.sa_mask);
	(void)sigaddset(&act.sa_mask, SIGPIPE);
	(void)sigaddset(&act.sa_mask, SIGALRM);
# ifdef        SIGWINCH
	(void)sigaddset(&act.sa_mask, SIGWINCH);
	(void)sigaction(SIGWINCH, &act, NULL);
# endif
	(void)sigaction(SIGPIPE, &act, NULL);
	act.sa_handler = dummy;
	(void)sigaction(SIGALRM, &act, NULL);
	act.sa_handler = s_rehash;
	(void)sigemptyset(&act.sa_mask);
	(void)sigaddset(&act.sa_mask, SIGHUP);
	(void)sigaction(SIGHUP, &act, NULL);
	act.sa_handler = s_restart;
	(void)sigaddset(&act.sa_mask, SIGINT);
	(void)sigaction(SIGINT, &act, NULL);
	act.sa_handler = s_die;
	(void)sigaddset(&act.sa_mask, SIGTERM);
	(void)sigaction(SIGTERM, &act, NULL);

#else
# ifndef HAVE_RELIABLE_SIGNALS
	(void)signal(SIGPIPE, dummy);
#  ifdef SIGWINCH
	(void)signal(SIGWINCH, dummy);
#  endif
# else
#  ifdef SIGWINCH
	(void)signal(SIGWINCH, SIG_IGN);
#  endif
	(void)signal(SIGPIPE, SIG_IGN);
# endif
	(void)signal(SIGALRM, dummy);
	(void)signal(SIGHUP, s_rehash);
	(void)signal(SIGTERM, s_die);
	(void)signal(SIGINT, s_restart);
#endif

#ifdef RESTARTING_SYSTEMCALLS
	/*
	** At least on Apollo sr10.1 it seems continuing system calls
	** after signal is the default. The following 'siginterrupt'
	** should change that default to interrupting calls.
	*/
	(void)siginterrupt(SIGALRM, 1);
#endif
}
#endif // !_Win32
