/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef lint
static char rcsid[] = "$Id: s_serv.c,v 1.26 2002/07/11 17:30:43 hex Exp $";
#endif

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#if defined(PCS) || defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
#include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#ifndef _WIN32
#include <utmp.h>
#else
#include <io.h>
#endif
#include <time.h>
#include "h.h"
#include "zlink.h"
#include "md5.h"
#include "unicode.h"

/* clean me!  condition added while moving win32 version from static
 * libc to dynamic msvcrt.dll;
 */
#ifndef _WIN32
extern char *strrchr(const char *s, int c);
#endif

static char buf[BUFSIZE];

/*
** m_functions execute protocol messages on this server:
**
**     cptr    is always NON-NULL, pointing to a *LOCAL* client
**             structure (with an open socket connected!). This
**             identifies the physical socket where the message
**             originated (or which caused the m_function to be
**             executed--some m_functions may call others...).
**
**     sptr    is the source of the message, defined by the
**             prefix part of the message if present. If not
**             or prefix not found, then sptr==cptr.
**
**             (!IsServer(cptr)) => (cptr == sptr), because
**             prefixes are taken *only* from servers...
**
**             (IsServer(cptr))
**                     (sptr == cptr) => the message didn't
**                     have the prefix.
**
**                     (sptr != cptr && IsServer(sptr) means
**                     the prefix specified servername. (?)
**
**                     (sptr != cptr && !IsServer(sptr) means
**                     that message originated from a remote
**                     user (not local).
**
**             combining
**
**             (!IsServer(sptr)) means that, sptr can safely
**             taken as defining the target structure of the
**             message in this server.
**
**     *Always* true (if 'parse' and others are working correct):
**
**     1)      sptr->from == cptr  (note: cptr->from == cptr)
**
**     2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**             *cannot* be a local connection, unless it's
**             actually cptr!). [MyConnect(x) should probably
**             be defined as (x == x->from) --msa ]
**
**     parc    number of variable parameter strings (if zero,
**             parv is allowed to be NULL)
**
**     parv    a NULL terminated list of parameter pointers,
**
**                     parv[0], sender (prefix string), if not present
**                             this points to an empty string.
**                     parv[1]...parv[parc-1]
**                             pointers to additional parameters
**                     parv[parc] == NULL, *always*
**
**             note:   it is guaranteed that parv[0]..parv[parc-1] are all
**                     non-NULL pointers.
*/


// m_server
//     parv[0] = sender prefix
//     parv[1] = servername
//     parv[2] = serverinfo/hopcount
//      parv[3] = serverinfo

#ifndef _WIN32
extern char *inet_ntoa(struct in_addr);
#endif

int m_server(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	int i, hop, ts = 0;
	char *ch, info[REALLEN + 1], *inpath, *host, *encr;
	aClient *acptr, *bcptr;
	struct ConfItem *aconf, *cconf;

	info[0] = '\0';
	inpath = get_client_name(cptr, FALSE);
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(cptr, "ERROR :No servername");
		return 0;
	}

	hop = 0;
	host = parv[1];

	if (parc > 4) {
		ts = atoi(parv[3]);
		hop = atoi(parv[2]);
		strncpyzt(info, parv[4], REALLEN);
	}
	else if (parc > 3 && atoi(parv[2])) {
		hop = atoi(parv[2]);
		strncpyzt(info, parv[3], REALLEN);
	}
	else if (parc > 2) {
		strncpyzt(info, parv[2], REALLEN);
		if (parc > 3 && ((i = strlen(info)) < (REALLEN - 2))) {
			strcat(info, " ");
			strncat(info, parv[3], REALLEN - i - 2);
			info[REALLEN] = '\0';
		}
	}
	// Check for "FRENCH " infection ;-) (actually this should
	// be replaced with routine to check the hostname syntax in
	// general). [ This check is still needed, even after the parse
	// is fixed, because someone can send "SERVER :foo bar " ].
	// Also, changed to check other "difficult" characters, now
	// that parse lets all through... --msa
	if (strlen(host) > HOSTLEN)
		host[HOSTLEN] = '\0';

	for (ch = host; *ch; ch++)
		if (*ch <= ' ' || *ch > '~')
			break;

	if (*ch || !strchr(host, '.')) {
		sendto_one(sptr, "ERROR :Bogus server name (%s)", sptr->name, host);
		sendto_ops("Bogus server name (%s) from %s", host, get_client_name(cptr, TRUE));
		sptr->since += 7;
		return 0;
	}

	if (IsPerson(cptr)) {
		// A local link that has been identified as a USER
		// tries something fishy... ;-)
		sendto_one(cptr, err_str(ERR_ALREADYREGISTRED), me.name, parv[0]);
		sendto_one(cptr, ":%s NOTICE %s :Sorry, but your IRC program doesn't appear to support changing servers.", me.name, cptr->name);
		sendto_ops("User %s trying to become a server %s", get_client_name(cptr, TRUE), host);
		sptr->since += 7;
		return 0;
	}

	// *WHEN* can it be that "cptr != sptr" ????? --msa
	// When SERVER command (like now) has prefix. -avalon

	// Get a pre-peek at the password...
	if (IsUnknown(cptr)) {
		// Borg88
		*cptr->vhost = 0;
		strcpy(cptr->sockhost, inet_ntoa(cptr->ip));
		aconf = find_conf_servern(host);
		if (!aconf) {
			sendto_locfailops("Access denied (No N line) %s", inpath);
			return exit_client(cptr, cptr, cptr, "No N line");
		}

		if (*aconf->passwd && !((*aconf->passwd == '+' && strlen(aconf->passwd) == 33 && !strcasecmp(aconf->passwd, MDPassword(cptr->passwd))) || ((*aconf->passwd != '+' || strlen(aconf->passwd) != 33) && !strcasecmp(aconf->passwd, cptr->passwd)))) {
			sendto_one(cptr, "ERROR :No Access (passwd mismatch) %s", inpath);
			sendto_locfailops("Access denied (passwd mismatch) %s", inpath);
			return exit_client(cptr, cptr, cptr, "Bad Password");
		}
		/* bzero(cptr->passwd, sizeof(cptr->passwd)); */
	}

	if ((acptr = find_name(host, NULL))) {
		aClient *ocptr;

		// This link is trying feed me a server that I already have
		// access through another path -- multiple paths not accepted
		// currently, kill this link immeatedly!!
		// Rather than KILL the link which introduced it, KILL the
		// youngest of the two links. -avalon
		acptr = acptr->from;
		ocptr = (cptr->firsttime > acptr->firsttime) ? acptr : cptr;
		acptr = (cptr->firsttime > acptr->firsttime) ? cptr : acptr;
		sendto_one(acptr, "ERROR :Server %s already exists from %s", host, (ocptr->from ? ocptr->from->name : "<nobody>"));
		sendto_ops("Link %s cancelled, server %s already exists from %s", get_client_name(acptr, TRUE), host, (ocptr->from ? ocptr->from->name : "<nobody>"));
		return exit_client(acptr, acptr, acptr, "Server Exists");
	}

	if ((acptr = find_client(host, NULL))) {
		// Server trying to use the same name as a person. Would
		// cause a fair bit of confusion. Enough to make it hellish
		// for a while and servers to send stuff to the wrong place.
		sendto_one(cptr, "ERROR :Nickname %s already exists!", host);
		sendto_locfailops("Link %s cancelled: Server/nick collision on %s", inpath, host);
		sendto_serv_butone(&me, ":%s GLOBOPS : Link %s cancelled: Server/nick collision on %s", parv[0], inpath, host);
		return exit_client(cptr, cptr, cptr, "Nick as Server");
	}

	if (IsServer(cptr)) {
		// Server is informing about a new server behind
		// this link. Create REMOTE server structure,
		// add it to list and propagate word to my other
		// server links...
		if (parc == 1 || info[0] == '\0') {
			sendto_one(cptr, "ERROR :No server info specified for %s", host);
			return 0;
		}

		// See if the newly found server is behind a guaranteed
		// leaf (L-line). If so, close the link.
		if ((aconf = find_conf_host(cptr->confs, host, CONF_LEAF)) && (!aconf->port || (hop > aconf->port))) {
			sendto_ops("Leaf-only link %s->%s - Closing", get_client_name(cptr, TRUE), aconf->host ? aconf->host : "*");
			sendto_one(cptr, "ERROR :Leaf-only link, sorry.");
			return exit_client(cptr, cptr, cptr, "Leaf Only");
		}

		if (!(aconf = find_conf_host(cptr->confs, host, CONF_HUB)) || (aconf->port && (hop > aconf->port))) {
			sendto_ops("Non-Hub link %s introduced %s(%s).", get_client_name(cptr, TRUE), host, aconf ? (aconf->host ? aconf->host : "*") : "!");
			return exit_client(cptr, cptr, cptr, "Too many servers");
		}

		// See if the newly found server has a Q line for it in
		// our conf. If it does, lose the link that brought it
		// into our network. Format:
		//
		// Q:<unused>:<reason>:<servername>
		//
		// Example: Q:*:for the hell of it:eris.Berkeley.EDU
		if ((aconf = find_conf_name(host, CONF_QUARANTINED_SERVER))) {
			sendto_ops_butone(NULL, &me, ":%s WALLOPS * :%s brought in %s, %s %s", me.name, get_client_name(cptr, FALSE), host, "closing link because", BadPtr(aconf->passwd) ? "reason unspecified" : aconf->passwd);
			sendto_one(cptr, "ERROR :%s is not welcome: %s. %s", host, BadPtr(aconf->passwd) ? "reason unspecified" : aconf->passwd, "Try another network");
			return exit_client(cptr, cptr, cptr, "Q-Lined Server");
		}

		acptr = make_client(cptr, find_server(parv[0], NULL));
		make_server(acptr);
		acptr->hopcount = hop;
		strncpyzt(acptr->name, host, sizeof(acptr->name));
		strncpyzt(acptr->info, info, sizeof(acptr->info));
		strncpyzt(acptr->serv->up, parv[0], sizeof(acptr->serv->up));
		SetServer(acptr);
		acptr->flags |= FLAGS_TS8;
		add_client_to_list(acptr);
		add_to_client_hash_table(acptr->name, acptr);

		// Old sendto_serv_but_one() call removed because we now
		// need to send different names to different servers
		// (domain name matching)
		for (i = 0; i <= highest_fd; i++) {
			if (!(bcptr = local[i]) || !IsServer(bcptr) || bcptr == cptr || IsMe(bcptr))
				continue;

			if (!(aconf = bcptr->serv->nline)) {
				sendto_ops("Lost N-line for %s on %s. Closing", get_client_name(cptr, TRUE), host);
				return exit_client(cptr, cptr, cptr, "Lost N line");
			}

			if (match(my_name_for_link(me.name, aconf), acptr->name) == 0)
				continue;

			if (ts)
				sendto_one(bcptr, ":%s SERVER %s %d %d :%s", parv[0], acptr->name, hop + 1, ts, acptr->info);
			else
				sendto_one(bcptr, ":%s SERVER %s %d :%s", parv[0], acptr->name, hop + 1, acptr->info);
		}
		// Check for U-line status -- Barubary
		if (find_conf_host(cptr->confs, acptr->name, CONF_UWORLD))
			acptr->flags |= FLAGS_ULINE;
		return 0;
	}

	if (!IsUnknown(cptr) && !IsHandshake(cptr))
		return 0;

	// A local link that is still in undefined state wants
	// to be a SERVER. Check if this is allowed and change
	// status accordingly...

	strncpyzt(cptr->name, host, sizeof(cptr->name));
	strncpyzt(cptr->info, info[0] ? info : me.name, sizeof(cptr->info));
	cptr->hopcount = hop;

	// check connection rules
	for (cconf = conf; cconf; cconf = cconf->next)
		if ((cconf->status == CONF_CRULEALL) && (match(cconf->host, host) == 0))
			if (crule_eval(cconf->passwd)) {
				ircstp->is_ref++;
				sendto_ops("Refused connection from %s.", get_client_host(cptr));
				return exit_client(cptr, cptr, cptr, "Disallowed by connection rule");
			}

	switch (check_server_init(cptr)) {
	case 0:
		return m_server_estab(cptr);

	case 1:
		sendto_ops("Access check for %s in progress", get_client_name(cptr, TRUE));
		return 1;

	default:
		ircstp->is_ref++;
		sendto_ops("Received unauthorized connection from %s.", get_client_host(cptr));
		sendto_serv_butone(&me, ":%s GLOBOPS :Recieved unauthorized connection from %s.", parv[0], get_client_host(cptr));
		return exit_client(cptr, cptr, cptr, "No C/N conf lines");
	}

}

int m_server_estab(aClient *cptr)
{
	aClient *acptr;
	struct ConfItem *aconf, *bconf;
	char *inpath, *host, *s;
	int i;

	inpath = get_client_name(cptr, TRUE); /* "refresh" inpath with host */
	host = cptr->name;

	if (!(aconf = find_conf(cptr->confs, host, CONF_NOCONNECT_SERVER))) {
		ircstp->is_ref++;
		sendto_one(cptr, "ERROR :Access denied. No N line for server %s", inpath);
		sendto_ops("Access denied. No N line for server %s", inpath);
		return exit_client(cptr, cptr, cptr, "No N line for server");
	}

	if (!(bconf = find_conf(cptr->confs, host, CONF_CONNECT_SERVER))) {
		ircstp->is_ref++;
		sendto_one(cptr, "ERROR :Only N (no C) field for server %s", inpath);
		sendto_ops("Only N (no C) field for server %s", inpath);
		return exit_client(cptr, cptr, cptr, "No C line for server");
	}

	if (!(*aconf->passwd == '+' && strlen(aconf->passwd) == 33 && StrEq(aconf->passwd + 1, MDPassword(cptr->passwd))) && !((*aconf->passwd != '+' || strlen(aconf->passwd) != 33) && StrEq(aconf->passwd, cptr->passwd))) {
		ircstp->is_ref++;
		sendto_one(cptr, "ERROR :No Access (passwd mismatch) %s", inpath);
		sendto_ops("Access denied (passwd mismatch) %s", inpath);
		return exit_client(cptr, cptr, cptr, "Bad Password");
	}

	bzero(cptr->passwd, sizeof(cptr->passwd));

#ifndef HUB
	for (i = 0; i <= highest_fd; i++) {
		if (local[i] && IsServer(local[i])) {
			ircstp->is_ref++;
			sendto_one(cptr, "ERROR :I'm a leaf not a hub");
			return exit_client(cptr, cptr, cptr, "I'm a leaf");
		}
	}
#endif

	if (IsUnknown(cptr)) {
		if (bconf->passwd[0])
			sendto_one(cptr, "PASS :%s", bconf->passwd);
		/* Pass my info to the new server */
		sendto_one(cptr, "SERVER %s 1 :%s", my_name_for_link(me.name, aconf), (me.info[0]) ? (me.info) : "IRCers United");
		send_protoctl(cptr);
	}
	else {
		s = strchr(aconf->host, '@');
		*s = '\0'; /* should never be NULL */
		Debug((DEBUG_INFO, "Check Usernames [%s]vs[%s]", aconf->host, cptr->username));
		if (match(aconf->host, cptr->username)) {
			*s = '@';
			ircstp->is_ref++;
			sendto_ops("Username mismatch [%s]v[%s] : %s", aconf->host, cptr->username, get_client_name(cptr, TRUE));
			sendto_one(cptr, "ERROR :No Username Match");
			return exit_client(cptr, cptr, cptr, "Bad User");
		}
		*s = '@';
	}

	det_confs_butmask(cptr, CONF_LEAF | CONF_HUB | CONF_NOCONNECT_SERVER | CONF_UWORLD);

	/* *WARNING*
	   In the following code in place of plain server's name we send what is returned
	   by get_client_name which may add the "sockhost" after the name. It's *very*
	   *important* that there is a SPACE between the name and sockhost (if present).
	   The receiving server will start the information field from this first blank
	   and thus puts the sockhost into info. ...a bit tricky, but you have been
	   warned, besides code is more neat this way...  --msa */
	SetServer(cptr);
	if (find_conf_host(cptr->confs, cptr->name, CONF_UWORLD))
		cptr->flags |= FLAGS_ULINE;
	cptr->flags |= FLAGS_TS8;
	nextping = time(NULL);

	add_to_client_hash_table(cptr->name, cptr);
	make_server(cptr);
	strcpy(cptr->serv->up, me.name);
	cptr->srvptr = &me;
	cptr->serv->nline = aconf;
	if (find_conf_host(cptr->confs, cptr->name, CONF_UWORLD))
		cptr->flags |= FLAGS_ULINE;

	*cptr->vhost = 0;

	/* synchronize now, if already got PROTOCTL or this is a service */
	return GotProtoctl(cptr) || (cptr->flags & FLAGS_ULINE)
		? m_server_burst(cptr) : 0;
}

int m_server_burst(aClient *cptr)
{
	int i;
	aClient *acptr;
	struct ConfItem *aconf;
	unsigned long insiz1, outsiz1, insiz2, outsiz2;
	double ratio1, ratio2;

#ifdef HUB
	sendto_serv_butone(&me, ":%s GLOBOPS :%s with %s established.", me.name, ZipOut(cptr) ? "Compressed link" : "Link", get_client_name(cptr, TRUE));
#endif

	sendto_locfailops("%s with %s established.", ZipOut(cptr) ? "Compressed link" : "Link", get_client_name(cptr, TRUE));

	SetOutBurst(cptr);

	if (ZipOut(cptr))
		zip_out_get_stats(cptr->serv->zip_out, &insiz1, &outsiz1, &ratio1);

	for (i = 0; i <= highest_fd; i++) {
		if (!(acptr = local[i]) || !IsServer(acptr) || acptr == cptr || IsMe(acptr))
			continue;
		if ((aconf = acptr->serv->nline) && !match(my_name_for_link(me.name, aconf), cptr->name))
			continue;
		sendto_one(acptr, ":%s SERVER %s 2 :%s", me.name, cptr->name, cptr->info);
	}

	/* Pass on my client information to the new server

	   First, pass only servers (idea is that if the link gets
	   cancelled beacause the server was already there,
	   there are no NICK's to be cancelled...). Of course,
	   if cancellation occurs, all this info is sent anyway,
	   and I guess the link dies when a read is attempted...? --msa

	   Note: Link cancellation to occur at this point means
	   that at least two servers from my fragment are building
	   up connection this other fragment at the same time, it's
	   a race condition, not the normal way of operation...

	   ALSO NOTE: using the get_client_name for server names--
	   see previous *WARNING*!!! (Also, original inpath
	   is destroyed...) */

	aconf = cptr->serv->nline;
	for (acptr = &me; acptr; acptr = acptr->prev) {
		/* acptr->from == acptr for acptr == cptr */
		if (acptr->from == cptr)
			continue;
		if (IsServer(acptr)) {
			if (match(my_name_for_link(me.name, aconf), acptr->name) == 0)
				continue;
			sendto_one(cptr, ":%s SERVER %s %d :%s", acptr->serv->up, acptr->name, acptr->hopcount + 1, acptr->info);
		}
	}

	for (acptr = &me; acptr; acptr = acptr->prev) {
		/* acptr->from == acptr for acptr == cptr */
		if (acptr->from == cptr)
			continue;
		if (IsPerson(acptr)) {
			/* IsPerson(x) is true only when IsClient(x) is true.
			   These are only true when *BOTH* NICK and USER have
			   been received. -avalon
			   Apparently USER command was forgotten... -Donwulff */
			sendto_one(cptr, "%s %s %d %d %s %s %s %lu :%s", (IsToken(cptr) ? TOK_NICK : MSG_NICK), acptr->name, acptr->hopcount + 1, acptr->lastnick, acptr->user->username, acptr->user->host, acptr->srvptr->name, acptr->user->servicestamp, acptr->info);

			send_umode(cptr, acptr, 0, SEND_UMODES, buf);

			send_user_joins(cptr, acptr);
		}
		else if (IsService(acptr)) {
			sendto_one(cptr, "NICK %s :%d", acptr->name, acptr->hopcount + 1);
			sendto_one(cptr, ":%s SERVICE * * :%s", acptr->name, acptr->info);
		}
	}

	/* Last, pass all channels plus statuses */
	{
		aChannel *chptr;

		for (chptr = channel; chptr; chptr = chptr->nextch) {
			send_channel_modes(cptr, chptr);
			if (chptr->topic_time)
				sendto_one(cptr, ":%s %s %s %s %lu :%s", me.name, (IsToken(cptr) ? TOK_TOPIC : MSG_TOPIC), chptr->chname, chptr->topic_nick, chptr->topic_time, chptr->topic);
		}
	}

	/* Pass on all services based q-lines */
	{
		aSqlineItem *tmp;

		for (tmp = sqline; tmp; tmp = tmp->next) {
			if (tmp->status != CONF_ILLEGAL)
				if (tmp->reason)
					sendto_one(cptr, ":%s SQLINE %s :%s", me.name, tmp->sqline, tmp->reason);
				else
					sendto_one(cptr, ":%s SQLINE %s", me.name, tmp->sqline);
		}
	}

	if (ZipOut(cptr)) {
		zip_out_get_stats(cptr->serv->zip_out, &insiz2, &outsiz2, &ratio2);
	}

	return 0;
}

/*
 * RPL_NOWON   - Online at the moment (Succesfully added to WATCH-list)
 * RPL_NOWOFF  - Offline at the moement (Succesfully added to WATCH-list)
 * RPL_WATCHOFF        - Succesfully removed from WATCH-list.
 * ERR_TOOMANYWATCH - Take a guess :>  Too many WATCH entries.
 */
static void show_watch(aClient * cptr, char *name, int rpl1, int rpl2)
{
	aClient *acptr;

	if ((acptr = find_person(name, NULL)))
		sendto_one(cptr, rpl_str(rpl1), me.name, cptr->name, acptr->name, acptr->user->username, IsHMask(acptr) ? acptr->user->vhost : acptr->user->host, acptr->lastnick);
	else
		sendto_one(cptr, rpl_str(rpl2), me.name, cptr->name, name, "*", "*", 0);
}

int m_watch(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	int len;
	char namebuf[USERLEN + HOSTLEN + 4], *s, **pav = parv, *user, *p = NULL, *def = "l";
	aClient *acptr;

	if (check_registered(sptr))
		return 0;

	if (parc < 2) {
		parc = 2;
		parv[1] = def;
	}

	for (s = (char *) strtoken(&p, *++pav, " "); s; s = (char *) strtoken(&p, NULL, " ")) {
		if ((user = strchr(s, '!')))
			*user++ = '\0'; // Not used

		// Prefix of "+", they want to add a name to their WATCH list.
		if (*s == '+') {
			if (do_nick_name(s + 1)) {
				if (sptr->notifies >= MAXWATCH) {
					sendto_one(sptr, err_str(ERR_TOOMANYWATCH), me.name, cptr->name, s + 1);
					continue;
				}
				add_to_notify_hash_table(s + 1, sptr);
			}

			show_watch(sptr, s + 1, RPL_NOWON, RPL_NOWOFF);
			continue;
		}

		// Prefix of "-", coward wants to remove somebody from their
		// WATCH list.  So do it. :-)
		if (*s == '-') {
			del_from_notify_hash_table(s + 1, sptr);
			show_watch(sptr, s + 1, RPL_WATCHOFF, RPL_WATCHOFF);
			continue;
		}

		// Fancy "C" or "c", they want to nuke their WATCH list and start
		// over, so be it.
		if (*s == 'C' || *s == 'c') {
			hash_del_notify_list(sptr);
			continue;
		}

		// Now comes the fun stuff, "S" or "s" returns a status report of
		// their WATCH list.  I imagine this could be CPU intensive if its
		// done alot, perhaps an auto-lag on this?
		if (*s == 'S' || *s == 's') {
			Link *lp;
			aNotify *anptr;
			int count = 0;

			// Send a list of how many users they have on their WATCH list
			// and how many WATCH lists they are on.
			anptr = hash_get_notify(sptr->name);
			if (anptr)
				for (lp = anptr->notify, count = 1; (lp = lp->next); count++);
			sendto_one(sptr, rpl_str(RPL_WATCHSTAT), me.name, parv[0], sptr->notifies, count);

			// Send a list of everybody in their WATCH list. Be careful
			// not to buffer overflow.
			if ((lp = sptr->notify) == NULL) {
				sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
				continue;
			}
			*buf = '\0';
			strcpy(buf, lp->value.nptr->nick);
			count = strlen(parv[0]) + strlen(me.name) + 10 + strlen(buf);
			while ((lp = lp->next)) {
				if (count + strlen(lp->value.nptr->nick) + 1 > BUFSIZE - 2) {
					sendto_one(sptr, rpl_str(RPL_WATCHLIST), me.name, parv[0], buf);
					*buf = '\0';
					count = strlen(parv[0]) + strlen(me.name) + 10;
				}
				strcat(buf, " ");
				strcat(buf, lp->value.nptr->nick);
				count += (strlen(lp->value.nptr->nick) + 1);
			}
			sendto_one(sptr, rpl_str(RPL_WATCHLIST), me.name, parv[0], buf);
			sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
			continue;
		}

		// Well that was fun, NOT.  Now they want a list of everybody in
		// their WATCH list AND if they are online or offline? Sheesh,
		// greedy arn't we?
		if (*s == 'L' || *s == 'l') {
			Link *lp = sptr->notify;

			while (lp) {
				if ((acptr = find_person(lp->value.nptr->nick, NULL)))
					sendto_one(sptr, rpl_str(RPL_NOWON), me.name, parv[0], acptr->name, acptr->user->username, acptr->user->host, acptr->lastnick);
				// But actually, only show them offline if its a capital
				// 'L' (full list wanted).
				else if (isupper(*s))
					sendto_one(sptr, rpl_str(RPL_NOWOFF), me.name, parv[0], lp->value.nptr->nick, "*", "*", lp->value.nptr->lasttime);
				lp = lp->next;
			}

			sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
			continue;
		}
	}

	return 0;
}

/*
** m_stats
**     parv[0] = sender prefix
**     parv[1] = statistics selector (defaults to Message frequency)
**     parv[2] = server name (current server defaulted, if omitted)
**
**     Currently supported are:
**             M = Message frequency (the old stat behaviour)
**             L = Local Link statistics
**             C = Report C and N configuration lines
*/
/*
** m_stats/stats_conf
**    Report N/C-configuration lines from this server. This could
**    report other configuration lines too, but converting the
**    status back to "char" is a bit akward--not worth the code
**    it needs...
**
**    Note:   The info is reported in the order the server uses
**            it--not reversed as in ircd.conf!
*/

static int report_array[17][3] = {
	{CONF_CONNECT_SERVER, RPL_STATSCLINE, 'C'},
	{CONF_NOCONNECT_SERVER, RPL_STATSNLINE, 'N'},
	{CONF_CLIENT, RPL_STATSILINE, 'I'},
	{CONF_KILL, RPL_STATSKLINE, 'K'},
	{CONF_ZAP, RPL_STATSKLINE, 'Z'},
	{CONF_QUARANTINED_NICK, RPL_STATSQLINE, 'Q'},
	{CONF_LEAF, RPL_STATSLLINE, 'L'},
	{CONF_OPERATOR, RPL_STATSOLINE, 'O'},
	{CONF_HUB, RPL_STATSHLINE, 'H'},
	{CONF_LOCOP, RPL_STATSOLINE, 'o'},
	{CONF_CRULEALL, RPL_STATSDLINE, 'D'},
	{CONF_CRULEAUTO, RPL_STATSDLINE, 'd'},
	{CONF_SERVICE, RPL_STATSSLINE, 'S'},
	{CONF_UWORLD, RPL_STATSULINE, 'U'},
	{CONF_MISSING, RPL_STATSXLINE, 'X'},
	{0, 0}
};

static void report_sqlined_nicks(sptr)
aClient *sptr;
{
	aSqlineItem *tmp;
	char *nickmask, *reason;

	for (tmp = sqline; tmp; tmp = tmp->next) {
		if (tmp->status != CONF_ILLEGAL) {
			nickmask = BadPtr(tmp->sqline) ? "<NULL>" : tmp->sqline;
			reason = BadPtr(tmp->reason) ? "<NULL>" : tmp->reason;
			sendto_one(sptr, rpl_str(RPL_SQLINE_NICK), me.name, sptr->name, nickmask, reason);
		}
	}
}

static void report_configured_links(aClient *sptr, int mask)
{
	static char null[] = "<NULL>";
	struct ConfItem *tmp;
	int *p, port, tmpmask;
	char c, *host, *pass, *name;

	tmpmask = (mask == CONF_MISSING) ? CONF_CONNECT_SERVER : mask;

	for (tmp = conf; tmp; tmp = tmp->next)
		if (tmp->status & tmpmask) {
			for (p = &report_array[0][0]; *p; p += 3)
				if (*p == tmp->status)
					break;
			if (!*p)
				continue;
			c = (char) *(p + 2);
			host = BadPtr(tmp->host) ? null : tmp->host;
			pass = BadPtr(tmp->passwd) ? null : tmp->passwd;
			name = BadPtr(tmp->name) ? null : tmp->name;
			port = (int) tmp->port;
			/*
			 * On K line the passwd contents can be
			 * displayed on STATS reply.    -Vesa
			 */
			// Same with Z-lines and q/Q-lines -- Barubary
			if ((tmp->status == CONF_KILL) || (tmp->status & CONF_QUARANTINE) || (tmp->status == CONF_ZAP)) {
/* These mods are to tell the difference between the different kinds
 * of klines.  the only effect it has is in the display.  --Russell
 */
// Now translates spaces to _'s to show comments in klines -- Barubary
				char *temp;

				if (!pass)
					strcpy(buf, "<NULL>");
				else {
					strcpy(buf, pass);
					for (temp = buf; *temp; temp++)
						if (*temp == ' ')
							*temp = '_';
				}
				// semicolon intentional -- Barubary
				if (tmp->status == CONF_QUARANTINED_NICK);
				// Hide password for servers -- Barubary
				else if (tmp->status & CONF_QUARANTINE)
					strcpy(buf, "*");
				else {
/* This wasn't documented before - comments aren't displayed for akills
   because they are all the same. -- Barubary */
					if (tmp->tmpconf == KLINE_AKILL)
						strcpy(buf, "*");
					/* KLINE_PERM == 0 - watch out when doing Z-lines. --
					   Barubary */
					if (tmp->status != CONF_ZAP) {
						if (tmp->tmpconf == KLINE_PERM)
							c = 'K';
						if (tmp->tmpconf == KLINE_TEMP)
							c = 'k';
						if (tmp->tmpconf == KLINE_AKILL)
							c = 'A';
					}
					else {
						if (tmp->tmpconf == KLINE_PERM)
							c = 'Z';
						if (tmp->tmpconf == KLINE_TEMP)
							c = 'z';
						if (tmp->tmpconf == KLINE_AKILL)
							c = 'S';
					}
				}
				sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host, buf, name, port, get_conf_class(tmp));
			}
			// connect rules are classless
			else if (tmp->status & CONF_CRULE)
				sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host, name);
			// Only display on X if server is missing
			else if (mask == CONF_MISSING) {
				if (!find_server(name, NULL))
					sendto_one(sptr, rpl_str(RPL_STATSXLINE), me.name, sptr->name, name, port);
			}
			else {
				if (!IsOper(sptr) && (mask & CONF_NOCONNECT_SERVER || mask & CONF_CONNECT_SERVER))
					sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, "*", name, port, get_conf_class(tmp));
				else

					sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host, name, port, get_conf_class(tmp));
			}
		}
	return;
}

// Used to blank out ports -- Barubary
char *get_client_name2(aClient * acptr, int showports)
{
	char *pointer = get_client_name(acptr, TRUE);

	if (!pointer)
		return NULL;
	if (showports)
		return pointer;
	if (!strrchr(pointer, '.'))
		return NULL;

	strcpy(strrchr(pointer, '.'), ".0]");

	return pointer;
}

int m_stats(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
#ifndef DEBUGMODE
	static char Sformat[] = ":%s %d %s SendQ SendM SendBytes RcveM RcveK Open_time :Idle";
	static char Lformat[] = ":%s %d %s %s %u %u %u %u %u %u :%u";
#else
	static char Sformat[] = ":%s %d %s SendQ SendM SendBytes RcveM RcveK Open_time CPU :Idle";
	static char Lformat[] = ":%s %d %s %s %u %u %u %u %u %u %s";
	char pbuf[96]; /* Should be enough for to ints */
#endif
	struct Message *mptr;
	aClient *acptr;
	char stat = parc > 1 ? parv[1][0] : '\0';
	int i;
	int doall = 0, wilds = 0, showports = IsAnOper(sptr), remote = 0;
	char *name;

	if (check_registered(sptr))
		return 0;

	if (MyClient(sptr) && ((!IsOper(sptr) && parc > 2) || (!IsAnOper(sptr) && stat != 'u'))) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (hunt_server(cptr, sptr, ":%s STATS %s :%s", 2, parc, parv) != HUNTED_ISME)
		return 0;

	if (parc > 2) {
		name = parv[2];
		if (!mycmp(name, me.name))
			doall = 2;
		else if (match(name, me.name) == 0)
			doall = 1;
		if (strchr(name, '*') || strchr(name, '?'))
			wilds = 1;
	}
	else
		name = me.name;

	switch (stat) {
	case 'L':
	case 'l':
		/* send info about connections which match, or all if the
		 * mask matches me.name.  Only restrictions are on those who
		 * are invisible not being visible to 'foreigners' who use
		 * a wild card based search to list it.
		 */
		sendto_one(sptr, Sformat, me.name, RPL_STATSLINKINFO, parv[0]);

		if (IsServer(cptr)) {
			remote = 1;
			wilds = 0;
		}

		for (i = 0; i <= highest_fd; i++) {
			if (!(acptr = local[i]))
				continue;
			if (IsInvisible(acptr) && (doall || wilds) && !(MyConnect(sptr) && IsOper(sptr)) && !IsAnOper(acptr) && (acptr != sptr))
				continue;
			if (remote && doall && !IsServer(acptr) && !IsMe(acptr))
				continue;
			if (remote && !doall && IsServer(acptr))
				continue;
			if (!doall && wilds && match(name, acptr->name))
				continue;
			if (!(parc == 2 && IsServer(acptr)) && !(doall || wilds) && mycmp(name, acptr->name))
				continue;
#ifdef DEBUGMODE
			sprintf(pbuf, "%d :%d", acptr->cputime, (acptr->user && MyConnect(acptr)) ? time(NULL) - acptr->user->last : 0);
#endif
			if (IsOper(sptr))
				sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO, parv[0], (isupper(stat)) ? get_client_name2(acptr, showports) : get_client_name(acptr, FALSE), (int) DBufLength(&acptr->sendQ), (int) acptr->sendM, (int) acptr->sendK, (int) acptr->receiveM, (int) acptr->receiveK, time(NULL) - acptr->firsttime,
#ifndef DEBUGMODE
						   (acptr->user && MyConnect(acptr)) ? time(NULL) - acptr->user->last : 0);
#else
						   pbuf);
#endif
			else if (!strchr(acptr->name, '.'))
				sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO, parv[0], (isupper(stat)) ? get_client_name2(acptr, showports) : get_client_name(acptr, FALSE), (int) DBufLength(&acptr->sendQ), (int) acptr->sendM, (int) acptr->sendK, (int) acptr->receiveM, (int) acptr->receiveK, time(NULL) - acptr->firsttime,
#ifndef DEBUGMODE
						   (acptr->user && MyConnect(acptr)) ? time(NULL) - acptr->user->last : 0);
#else
						   pbuf);
#endif
		}
		break;

	case 'C':
	case 'c':
		report_configured_links(sptr, CONF_CONNECT_SERVER | CONF_NOCONNECT_SERVER);
		break;

	case 'H':
	case 'h':
		report_configured_links(sptr, CONF_HUB | CONF_LEAF);
		break;

	case 'I':
	case 'i':
		report_configured_links(sptr, CONF_CLIENT);
		break;

	case 'K':
	case 'k':
		report_configured_links(sptr, CONF_KILL | CONF_ZAP);
		break;

	case 'M':
	case 'm':
		for (mptr = msgtab; mptr->cmd; mptr++)
			if (mptr->count)
#ifndef DEBUGMODE
				sendto_one(sptr, rpl_str(RPL_STATSCOMMANDS), me.name, parv[0], mptr->cmd, mptr->count, mptr->bytes);
#else
				sendto_one(sptr, rpl_str(RPL_STATSCOMMANDS), me.name, parv[0], mptr->cmd, mptr->count, mptr->bytes, mptr->lticks, mptr->lticks / CLOCKS_PER_SEC, mptr->rticks, mptr->rticks / CLOCKS_PER_SEC);
#endif
		break;

	case 'o':
	case 'O':
		if (IsAnOper(sptr))
			report_configured_links(sptr, CONF_OPS);
		break;

	case 'p':
	case 'P':
		{
			int i, utf, size;
			struct codepage *cp;

			sendto_one(sptr, ":%s %d %s :Name Alias Ref Bytes", me.name, RPL_STATSDEBUG, sptr->name);

			for (i = 0; i < sizeof(cps) / sizeof(cps[0]); ++i)
				if (cp = cps[i])
					sendto_one(sptr, rpl_str(RPL_STATSCODEPAGE), me.name, parv[0], cp->name, cp->alias[0] ? cp->alias : "-", cp->refcount, cp->bytes);

			for (i = 0, utf = 0; i < highest_fd; ++i)
				if (local[i] && IsClient(local[i]) && !local[i]->codepage)
					++utf;

			sendto_one(sptr, rpl_str(RPL_STATSCODEPAGE), me.name, parv[0], "utf8", "none", utf, 0);
			break;
		}

	case 'Q':
		report_configured_links(sptr, CONF_QUARANTINE);
		break;

	case 'q':
		report_sqlined_nicks(sptr);
		break;

	case 'R':
	case 'r':
#ifdef DEBUGMODE
		send_usage(sptr, parv[0]);
#endif
		break;

	case 'D':
		report_configured_links(sptr, CONF_CRULEALL);
		break;

	case 'd':
		report_configured_links(sptr, CONF_CRULE);
		break;

	case 'S':
	case 's':
		report_configured_links(sptr, CONF_SERVICE);
		break;

	case 'T':
	case 't':
		tstats(sptr, parv[0]);
		break;

	case 'U':
		report_configured_links(sptr, CONF_UWORLD);
		break;

	case 'u':
		{
			register time_t tmpnow;

			tmpnow = time(NULL) - me.since;
			sendto_one(sptr, rpl_str(RPL_STATSUPTIME), me.name, parv[0], tmpnow / 86400, (tmpnow / 3600) % 24, (tmpnow / 60) % 60, tmpnow % 60);
			break;
		}

	/* Borg88, V-Lines stats */
	case 'V':
	case 'v':
		vstats(sptr, parv[0]);
		break;

	case 'X':
	case 'x':
		report_configured_links(sptr, CONF_MISSING);
		break;

	case 'Y':
	case 'y':
		report_classes(sptr);
		break;

	case 'Z':
	case 'z':
		count_memory(sptr, parv[0]);
		break;

	default:
		stat = '*';
		break;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, parv[0], stat);
	return 0;
}

// Note: At least at protocol level ERROR has only one parameter,
// although this is called internally from other functions
// --msa
//
//     parv[0] = sender prefix
//     parv[*] = parameters
int m_error(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	char *para;

	para = (parc > 1 && *parv[1] != '\0') ? parv[1] : "<>";

	Debug((DEBUG_ERROR, "Received ERROR message from %s: %s", sptr->name, para));

	/*
	   ** Ignore error messages generated by normal user clients ** (because
	   ill-behaving user clients would flood opers ** screen otherwise). Pass
	   ERROR's from other sources to ** the local operator... */
	if (IsPerson(cptr) || IsUnknown(cptr) || IsService(cptr))
		return 0;

	if (cptr == sptr) {
		sendto_serv_butone(&me, ":%s GLOBOPS :ERROR from %s -- %s", me.name, get_client_name(cptr, FALSE), para);

		sendto_locfailops("ERROR :from %s -- %s", get_client_name(cptr, FALSE), para);
	}
	else {
		sendto_serv_butone(&me, ":%s GLOBOPS :ERROR from %s via %s -- %s", me.name, sptr->name, get_client_name(cptr, FALSE), para);

		sendto_ops("ERROR :from %s via %s -- %s", sptr->name, get_client_name(cptr, FALSE), para);
	}
	return 0;
}

Link *helpign = NULL;

/* Now just empty ignore-list, in future reload dynamic help.
 * Move out to help.c -Donwulff */
void reset_help()
{
	free_str_list(helpign);
}

// m_help -- deprecated, only advices local users to go to help channel
int m_help(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	if (!check_registered(sptr) && MyConnect(sptr) && IsPerson(sptr))
		sendto_one(sptr, ":%s 292 %s :Deprecated command. Please use /ns help, " "/cs help, /ms help or go to the network's official help channel " "(check /motd for that).", me.name, parv[0]);
	return 0;
}

/***********************************************************************
 * m_connect() - Added by Jto 11 Feb 1989
 ***********************************************************************/

/*
** m_connect
**     parv[0] = sender prefix
**     parv[1] = servername
**     parv[2] = port number
**     parv[3] = remote server
*/
int m_connect(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int parc;
char *parv[];
{
	int port, tmpport, retval;
	struct ConfItem *aconf, *cconf;
	aClient *acptr;

	if (check_registered(sptr))
		return 0;

	if (!IsPrivileged(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return -1;
	}

	/* Only allow LocOps to make local CONNECTS --SRB */
	if (MyClient(sptr) && !OPIsOper(sptr) && parc > 3) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (MyClient(sptr) && !OPIsAnOper(sptr) && parc <= 3) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (hunt_server(cptr, sptr, ":%s CONNECT %s %s :%s", 3, parc, parv) != HUNTED_ISME)
		return 0;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "CONNECT");
		return -1;
	}

	if ((acptr = find_server(parv[1], NULL))) {
		sendto_one(sptr, ":%s NOTICE %s :Connect: Server %s %s %s.", me.name, parv[0], parv[1], "already exists from", acptr->from->name);
		return 0;
	}

	for (aconf = conf; aconf; aconf = aconf->next)
		if (aconf->status == CONF_CONNECT_SERVER && match(parv[1], aconf->name) == 0)
			break;
	// Checked first servernames, then try hostnames.
	if (!aconf)
		for (aconf = conf; aconf; aconf = aconf->next)
			if (aconf->status == CONF_CONNECT_SERVER && (match(parv[1], aconf->host) == 0 || match(parv[1], strchr(aconf->host, '@') + 1) == 0))
				break;

	if (!aconf) {
		sendto_one(sptr, "NOTICE %s :Connect: Host %s not listed in ircd.conf", parv[0], parv[1]);
		return 0;
	}

	/* Get port number from user, if given. If not specified, use the
	   default form configuration structure. If missing from there, then use
	   the precompiled default. */
	tmpport = port = aconf->port;
	if (!BadPtr(parv[2])) {
		if ((port = atoi(parv[2])) <= 0) {
			sendto_one(sptr, "NOTICE %s :Connect: Illegal port number", parv[0]);
			return 0;
		}
	}

	/* Evaluate connection rules...  If no rules found, allow the
	   connect.  Otherwise stop with the first true rule (ie: rules are ored
	   together.  Oper connects are effected only by D lines (CRULEALL) not
	   d lines (CRULEAUTO). */
	for (cconf = conf; cconf; cconf = cconf->next)
		if ((cconf->status == CONF_CRULEALL) && (match(cconf->host, aconf->name) == 0))
			if (crule_eval(cconf->passwd)) {
				sendto_one(sptr, "NOTICE %s :Connect: Disallowed by rule: %s", parv[0], cconf->name);
				return 0;
			}

	// Notify all operators about remote connect requests
	if (!IsAnOper(cptr)) {
		sendto_serv_butone(&me, ":%s WALLOPS :Remote CONNECT %s %s from %s", me.name, parv[1], parv[2] ? parv[2] : "", get_client_name(sptr, FALSE));
#if defined(USE_SYSLOG) && defined(SYSLOG_CONNECT)
		syslog(LOG_DEBUG, "CONNECT From %s : %s %d", parv[0], parv[1], parv[2] ? parv[2] : "");
#endif
	}

	aconf->port = port;
	switch (retval = connect_server(aconf, sptr, NULL)) {
		case 0:
			sendto_one(sptr, ":%s NOTICE %s :*** Connecting to "
				"%s[%s:%d].", me.name, parv[0], aconf->name,
				aconf->host, aconf->port);
			break;
		case -1:
			sendto_one(sptr, ":%s NOTICE %s :*** Couldn't connect "
				"to %s.", me.name, parv[0], aconf->host);
			break;
		case -2:
			sendto_one(sptr, ":%s NOTICE %s :*** Host %s is "
				"unknown.", me.name, parv[0], aconf->host);
			break;
		case -4:
			sendto_one(sptr, ":%s NOTICE %s :*** %s is not enabled "
				"for connection.", me.name, parv[0],
				aconf->name);
			break;
		default:
			sendto_one(sptr, ":%s NOTICE %s :*** Connection to %s "
				"failed: %s", me.name, parv[0], aconf->host,
				strerror(retval));
	}
	aconf->port = tmpport;
	return 0;
}

/* m_gnotice  (Russell) sort of like wallop, but only to +g clients on
** this server.
**     parv[0] = sender prefix
**     parv[1] = message text
*/
int m_gnotice(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	char *message, *pv[4];

	if (check_registered(sptr))
		return 0;

	message = parc > 1 ? parv[1] : NULL;

	if (BadPtr(message)) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GNOTICE");
		return 0;
	}

	if (!IsServer(sptr) && MyConnect(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GNOTICE :%s", parv[0], message);
	sendto_failops("from %s: %s", parv[0], message);
	return 0;
}

// m_svskill
//     parv[0] = servername
//     parv[1] = client
//     parv[2] = kill message
int m_svskill(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	aClient *acptr;
	char *comment = (parc > 2 && parv[2]) ? parv[2] : "SVS Killed";

	if (!IsULine(cptr, sptr))
		return -1;

	if (hunt_server(cptr, sptr, ":%s SVSKILL %s :%s", 1, parc, parv) != HUNTED_ISME)
		return 0;

	if (parc < 1 || (!(acptr = find_client(parv[1], NULL))))
		return 0;

	return exit_client(cptr, acptr, sptr, comment);
}

/*
** parv[1] - password *OR* reason if no X:line
** parv[2] - reason for restart (optional & only if X:line exists)
**
** The password is only valid if there is a matching X line in the
** config file. If it is not,  then it becomes the
*/
int m_restart(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	char *pass = NULL;

	if (check_registered(sptr))
		return 0;

	if (!MyClient(sptr) || !OPIsAdmin(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if ((pass = find_restartpass())) {
		if (parc < 2) {
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "RESTART");
			return 0;
		}

		if (strcmp(pass, parv[1])) {
			sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
			return 0;
		}

		/* Hack to make the code after this if { } easier: we assign the
		   comment to the first param, as if we had not had an X:line. We do
		   not need the password now anyways. Accordingly we decrement parc ;)
		   -- NikB */
		parv[1] = parv[2];
		parc--;
	}

#ifdef USE_SYSLOG
	syslog(LOG_WARNING, "Server RESTART by %s - %s\n", get_client_name(sptr, FALSE), (parc > 1 ? parv[1] : "No reason"));
#endif

	sendto_ops("Server is Restarting by request of %s", parv[0]);
	server_reboot((parv[1] ? parv[1] : "No Reason"));
	return 0;
}

/* parv[0] = sender prefix, parv[1] = servername */
int m_trace(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
	char *tname;
	time_t now;
	aClass *cltmp;
	aClient *acptr, *acptr2;
	int i, cnt = 0, wilds, dow, doall, link_s[MAXCONNECTIONS], link_u[MAXCONNECTIONS];

	if (check_registered(sptr))
		return 0;

	if (!IsPrivileged(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc > 2)
		if (hunt_server(cptr, sptr, ":%s TRACE %s :%s", 2, parc, parv))
			return 0;

	if (parc > 1)
		tname = parv[1];
	else
		tname = me.name;

	switch (hunt_server(cptr, sptr, ":%s TRACE :%s", 1, parc, parv)) {
		/* note: gets here only if parv[1] exists */
	case HUNTED_PASS:{
			aClient *ac2ptr;

			ac2ptr = next_client(client, tname);
			sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name, parv[0], version, debugmode, tname, ac2ptr->from->name);
			return 0;
		}

	case HUNTED_ISME:
		break;

	default:
		return 0;
	}

	doall = (parv[1] && (parc > 1)) ? !match(tname, me.name) : TRUE;
	wilds = !parv[1] || strchr(tname, '*') || strchr(tname, '?');
	dow = wilds || doall;

#ifndef _WIN32
	for (i = 0; i < MAXCONNECTIONS; i++)
		link_s[i] = 0, link_u[i] = 0;
#else
	bzero(link_s, sizeof(link_s));
	bzero(link_u, sizeof(link_u));
#endif

	if (doall)
		for (acptr = client; acptr; acptr = acptr->next)
			if (IsPerson(acptr))
				link_u[acptr->from->fd]++;
			else if (IsServer(acptr))
				link_s[acptr->from->fd]++;

	/* report all direct connections */

	now = time(NULL);
	for (i = 0; i <= highest_fd; i++) {
		char *name;
		int class;

		if (!(acptr = local[i])) /* Local Connection? */
			continue;

		/* More bits of code to allow oers to see all users on remote traces */
		if (!IsOper(sptr) && !IsAnOper(acptr) && (acptr != sptr))
			continue;

		if (!doall && wilds && match(tname, acptr->name))
			continue;

		if (!dow && mycmp(tname, acptr->name))
			continue;

		name = get_client_name(acptr, FALSE);
		class = get_client_class(acptr);

		switch (acptr->status) {
		case STAT_CONNECTING:
			sendto_one(sptr, rpl_str(RPL_TRACECONNECTING), me.name, parv[0], class, name);
			cnt++;
			break;

		case STAT_HANDSHAKE:
			sendto_one(sptr, rpl_str(RPL_TRACEHANDSHAKE), me.name, parv[0], class, name);
			cnt++;
			break;

		case STAT_ME:
			break;

		case STAT_UNKNOWN:
			sendto_one(sptr, rpl_str(RPL_TRACEUNKNOWN), me.name, parv[0], class, name);
			cnt++;
			break;

		case STAT_CLIENT:
			if (IsOper(sptr) || (IsAnOper(acptr) && !IsInvisible(acptr))) {
				if (IsAnOper(acptr))
					sendto_one(sptr, rpl_str(RPL_TRACEOPERATOR), me.name, parv[0], class, name, now - acptr->lasttime);
				else
					sendto_one(sptr, rpl_str(RPL_TRACEUSER), me.name, parv[0], class, name, now - acptr->lasttime);
				cnt++;
			}
			break;

		case STAT_SERVER:
			if (acptr->serv->user)
				sendto_one(sptr, rpl_str(RPL_TRACESERVER), me.name, parv[0], class, link_s[i], link_u[i], name, acptr->serv->by, acptr->serv->user->username, acptr->serv->user->host, now - acptr->lasttime);
			else
				sendto_one(sptr, rpl_str(RPL_TRACESERVER), me.name, parv[0], class, link_s[i], link_u[i], name, *(acptr->serv->by) ? acptr->serv->by : "*", "*", me.name, now - acptr->lasttime);
			cnt++;
			break;

		case STAT_SERVICE:
			sendto_one(sptr, rpl_str(RPL_TRACESERVICE), me.name, parv[0], class, name);
			cnt++;
			break;

		case STAT_LOG:
#ifdef DEBUGMODE /* maybe it's wrong to skip this if DEBUGMODE is not defined, but at least it doesn't break things - draky */
			sendto_one(sptr, rpl_str(RPL_TRACELOG), me.name, parv[0], debugfn, acptr->port);
#endif
			cnt++;
			break;

		default:        /* ...we actually shouldn't come here... --msa */
			sendto_one(sptr, rpl_str(RPL_TRACENEWTYPE), me.name, parv[0], name);
			cnt++;
			break;
		}
	}

	/* Add these lines to summarize the above which can get rather long and
	   messy when done remotely - Avalon */
	if (!IsAnOper(sptr) || !cnt) {
		if (cnt)
			return 0;

		/* let the user have some idea that its at the end of the trace */
		sendto_one(sptr, rpl_str(RPL_TRACESERVER), me.name, parv[0], 0, link_s[me.fd], link_u[me.fd], me.name, "*", "*", me.name);
		return 0;
	}

	for (cltmp = FirstClass(); doall && cltmp; cltmp = NextClass(cltmp))
		if (Links(cltmp) > 0)
			sendto_one(sptr, rpl_str(RPL_TRACECLASS), me.name, parv[0], Class(cltmp), Links(cltmp));
	return 0;
}
