/************************************************************************
 *   IRC - Internet Relay Chat, common/send.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                   University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef lint
static char rcsid[] = "$Id: send.c,v 1.26 2002/06/14 21:41:31 hex Exp $";
#endif

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>

#ifdef _WIN32
# include <io.h>
#endif

#include "zlink.h"
#include "send.h"
#include "unicode.h"

static char sendbuf[MAXMSGLEN];
static int send_message PROTO((aClient *, char *, int));

static int sentalong[MAXCONNECTIONS];

// dead_link
//         An error has been detected. The link *must* be closed,
//         but *cannot* call ExitClient (m_bye) from here.
//         Instead, mark it with FLAGS_DEADSOCKET. This should
//         generate ExitClient from the main loop.
//
//         If 'notice' is not NULL, it is assumed to be a format
//         for a message to local opers. I can contain only one
//         '%s', which will be replaced by the sockhost field of
//         the failing link.
//
//         Also, the notice is skipped for "uninteresting" cases,
//         like Persons and yet unknown connections...
static int dead_link(aClient *to, char *notice)
{
	to->flags |= FLAGS_DEADSOCKET;
	// If because of BUFFERPOOL problem then clean dbuf's now so that
	// notices don't hurt operators below.
	DBufClear(&to->recvQ);
	DBufClear(&to->sendQ);
	if (!IsPerson(to) && !IsUnknown(to) && !(to->flags & FLAGS_CLOSING))
		sendto_failops_whoare_opers(notice, get_client_name(to, FALSE));
	Debug((DEBUG_ERROR, notice, get_client_name(to, FALSE)));
	return -1;
}


// flush_connections
// Used to empty all output buffers for all connections. Should only
// be called once per scan of connections. There should be a select in
// here perhaps but that means either forcing a timeout or doing a poll.
// When flushing, all we do is empty the obuffer array for each local
// client and try to send it. if we cant send it, it goes into the sendQ
// -avalon
void flush_connections(int fd)
{
#ifdef SENDQ_ALWAYS
	int i;
	aClient *cptr;

	if (fd == me.fd) {
		for (i = highest_fd; i >= 0; i--)
			if ((cptr = local[i]) && DBufLength(&cptr->sendQ) > 0)
				send_queued(cptr);
	}
	else if (fd >= 0 && (cptr = local[fd]) && DBufLength(&cptr->sendQ) > 0)
		send_queued(cptr);
#endif
}

// send_message
//        Internal utility which delivers one message buffer to the
//        socket. Takes care of the error handling and buffering, if
//        needed.

static int send_message(aClient *to, char *msg, int len)
#ifdef SENDQ_ALWAYS
{
#ifdef DEBUGMODE
	char tmp[MAXMSGLEN];
	strncpyzt(tmp, msg, sizeof(tmp));
#endif

	if (IsDead(to))
		return 0; // This socket has already been marked as dead

	if (ZipOut(to)) {
		// only on netjoins
		// int ldata = (to->flags & FLAGS_BURST), oldlen = len;
		int ldata = IsOutBurst(to);

		msg = zip_output(to->serv->zip_out, msg, &len, 0, &ldata);
		if (len == -1) {
			sendto_realops("Zipout error for %s: (%d) %s\n", to->name, ldata, msg);
			return dead_link(to, "Zip output error for %s" /*, IRCERR_ZIP*/);
		}

#ifdef DEBUGMODE
	if (IsServer(to))
		sendto_syschan("raw_out", ZipOut(to) ? "-> Z: %s, %s" : "-> N: %s, %s", to->name, tmp);
#endif

		// buffer not filled yet, waiting for data or an explicit flush
		if (len == 0)
			return 0;
	}

	if (DBufLength(&to->sendQ) > get_sendq(to)) {
		if (IsServer(to))
			sendto_ops("Max SendQ limit exceeded for %s: %d > %d", get_client_name(to, FALSE), DBufLength(&to->sendQ), get_sendq(to));
			return dead_link(to, "Max Sendq exceeded");
		}
	else if (dbuf_put(&to->sendQ, msg, len) < 0)
		return dead_link(to, "Buffer allocation error for %s");

	// Update statistics. The following is slightly incorrect
	// because it counts messages even if queued, but bytes
	// only really sent. Queued bytes get updated in SendQueued.
	to->sendM += 1;
	me.sendM += 1;
	if (to->acpt != &me)
		to->acpt->sendM += 1;

	// This little bit is to stop the sendQ from growing too large when
	// there is no need for it to. Thus we call send_queued() every time
	// 2k has been added to the queue since the last non-fatal write.
	// Also stops us from deliberately building a large sendQ and then
	// trying to flood that link with data (possible during the net
	// relinking done by servers with a large load).
	if (DBufLength(&to->sendQ)/1024 > to->lastsq)
		send_queued(to);

#ifdef DEBUGMODE
	if (IsServer(to))
		sendto_syschan("raw_out", ZipOut(to) ? "-> Z: %s, %s" : "-> N: %s, %s", to->name, tmp);
#endif

	return 0;
}
#else
{
	int rlen = 0;

	if (IsDead(to))
		return 0; // This socket has already been marked as dead

	// DeliverIt can be called only if SendQ is empty...
	if ((DBufLength(&to->sendQ) == 0) && (rlen = deliver_it(to, msg, len)) < 0)
		return dead_link(to,"Write error to %s, closing link");
	else if (rlen < len) {
		// Was unable to transfer all of the requested data. Queue
		// up the remainder for some later time...
		if (DBufLength(&to->sendQ) > get_sendq(to)) {
			sendto_ops("Max SendQ limit exceeded for %s : %d > %d", get_client_name(to, FALSE), DBufLength(&to->sendQ), get_sendq(to));
			return dead_link(to, "Max Sendq exceeded");
		}
		else if (dbuf_put(&to->sendQ,msg+rlen,len-rlen) < 0)
			return dead_link(to,"Buffer allocation error for %s");
	}
	// Update statistics. The following is slightly incorrect
	// because it counts messages even if queued, but bytes
	// only really sent. Queued bytes get updated in SendQueued.
	to->sendM += 1;
	me.sendM += 1;
	if (to->acpt != &me)
		to->acpt->sendM += 1;
	return 0;
}
#endif

/* send_queued
** This function is called from the main select-loop (or whatever)
** when there is a chance the some output would be possible. This
** attempts to empty the send queue as far as possible...
*/
int send_queued(aClient *to)
{
	char *msg;
	int len, rlen, more_data = 0;

#ifndef pyr
	if (IsBlocked(to)) return -1; // Can't write to already blocked socket
#endif

	/* Once socket is marked dead, we cannot start writing to it,
	 * even if the error is removed... */
	if (IsDead(to)) {
#ifndef SENDQ_ALWAYS
		return dead_link(to, "send_queued called for a DEADSOCKET:%s");
#else
		return -1;
#endif
	}

	if (ZipOut(to) && zip_is_data_out(to->serv->zip_out)) {
		if (DBufLength(&to->sendQ))
			more_data = 1;
		else {
			int ldata = IsOutBurst(to);

			msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);
			if (len == -1) {
				sendto_syschan("links", "Zipout error for %s: (%d) %s\n", to->name, ldata, msg);
				return dead_link(to, "Zip output error for %s" /*, IRCERR_ZIP */);
			}

			/* silently stick this on the sendq... */
			if (!dbuf_put(&to->sendQ, msg, len))
				return dead_link(to, "Buffer allocation error for %s" /*, IRCERR_BUFALLOC */);
		}
	}

	while (DBufLength(&to->sendQ) > 0) {
		msg = dbuf_map(&to->sendQ, &len);

		/* Returns always len > 0 */
		if ((rlen = deliver_it(to, msg, len)) < 0)
			return dead_link(to,"Write error to %s, closing link");

		dbuf_delete(&to->sendQ, rlen);
		to->lastsq = DBufLength(&to->sendQ)/1024;
		if (rlen < len) {
			/* If we can't write full message, mark the socket as "blocking"
			 * and stop trying. -Donwulff */
			SetBlocked(to);
			break;
		}

		if (more_data && !DBufLength(&to->sendQ)) {
			int ldata = IsOutBurst(to);
			more_data = 0;
			msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);

			if (len == -1) {
				sendto_syschan("links", "Zipout error for %s: (%d) %s\n", to->name, ldata, msg);
				return dead_link(to, "Zip output error for %s" /*, IRCERR_ZIP */);
			}

			/* silently stick this on the sendq... */
			if (!dbuf_put(&to->sendQ, msg, len))
				return dead_link(to, "Buffer allocation error for %s" /*, IRCERR_BUFALLOC */);
		}
	}

	if (IsOutBurst(to)) {
		ResOutBurst(to);

		if (ZipOut(to)) {
			unsigned long insiz = 0, outsiz = 0;
			double ratio = 1;
			zip_out_get_stats(to->serv->zip_out, &insiz, &outsiz, &ratio);
			sendto_syschan("links", "NetBurst to %s complete, bytes sent: %d (%d zipped), ratio: %.2f%%", to->name, insiz, outsiz, ratio);
		} else {
			sendto_syschan("links", "NetBurst to %s complete, bytes sent: %d", to->name, to->sendK * 1024 + to->sendB);
		}
	}

	return (IsDead(to)) ? -1 : 0;
}

/* send message to single client */

static void vsendto_one(aClient *to, char *pattern, va_list vl) {
	char *translated;
#ifdef VMS
	extern int goodbye;
	if (StrEq("QUIT", pattern))
		goodbye = 1;
#endif

	vsprintf(sendbuf, pattern, vl);

	if (MyClient(to) && to->codepage && (translated = unicode_to_codepage(sendbuf, to->codepage - 1)) != sendbuf) {
		strncpy(sendbuf, translated, sizeof(sendbuf));
		sendbuf[sizeof(sendbuf) - 1] = '\0';
	}

	Debug((DEBUG_SEND,"Sending [%s] to %s", sendbuf, to->name));

	if (to->from)
		to = to->from;

	if (to->fd < 0) {
		Debug((DEBUG_ERROR, "Local socket %s with negative fd... AARGH!", to->name));
	}
	else if (IsMe(to)) {
		sendto_ops("Trying to send [%s] to myself!", sendbuf);
		return;
	}

	/* truncate long messages, stick clients to 512 bytes as declared by RFC */
	{
		char *limit = sendbuf + (MyClient(to) ? 512 : sizeof(sendbuf)) - 4;

		for (translated = sendbuf; *translated && translated < limit; ++translated);

		*translated++ = '\r';
		*translated++ = '\n';
		*translated++ = '\0';
	}

	send_message(to, sendbuf, strlen(sendbuf));
}


// sendto_prefix_one()
// to - destination client
// from - client which message is from
// NOTE: NEITHER OF THESE SHOULD//EVER* BE NULL!!
// -avalon

static void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
{
	char *p, *par = NULL;
	int flag = 0;
	struct User *user;
	static char temp[1024], sender[HOSTLEN+NICKLEN+USERLEN+5];

	par = va_arg(vl, char *);
	if (to && from && MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
		user = from->user;
		strcpy(sender, from->name);
		if (user) {
			if (*user->username) {
				strcat(sender, "!");
				strcat(sender, user->username);
			}

			/* local users disclosure, masking issue... -- hex */
			if (*user->host /*&& !MyConnect(from)*/) {
				strcat(sender, "@");
				strcat(sender, IsHMask(from) ? user->vhost : user->host);
				flag = 1;
			}
		}
		/* flag is used instead of index(sender, '@') for speed and
		   also since username/nick may have had a '@' in them. -avalon */
		if (!flag && MyConnect(from) && *user->host) {
			char *h;
			strcat(sender, "@");
			h = (*from->vhost) ? (IsHMask(from) ? from->user->vhost : from->user->host) : from->sockhost;
			strcat(sender, h);
		}
		par = sender;
	}

	/* since prefix is always :%s, copy it here */
	*temp = ':';
	strcpy(temp + 1, par);

	/* if we're sending to a non-op, make sure we don't reveal servers */
	if (!BadPtr(server_mask) && MyClient(to) && !IsOper(to) && !strchr(temp, '@'))
		/* let locops see their clients */
		if (!IsLocOp(to) || strcmp(temp + 1, me.name))
			strcpy(temp + 1, server_mask);

	/* append the rest of the format string */
	strcat(temp, pattern + 3);

	vsendto_one(to, temp, vl);
}


void sendto_one(aClient *to, char *pattern, ...)
{
	va_list vl;

	va_start(vl, pattern);
	vsendto_one(to, pattern, vl);
	va_end(vl);
}


static void vsendto_channel_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, va_list vl) {
	int i;
	Link *lp;
	aClient *acptr;

	bzero(sentalong, sizeof(sentalong));
	for (lp = chptr->members; lp; lp = lp->next) {
		acptr = lp->value.cptr;
		if (acptr->from == one || (lp->flags & CHFL_ZOMBIE))
			continue; /* ...was the one I should skip */
		i = acptr->from->fd;
		if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[i] = 1;
		}
		else {
			/* Now check whether a message has been sent to this remote link already */
			if (sentalong[i] == 0) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = 1;
			}
		}
	}
	return;
}

void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	va_list vl;
	va_start(vl, pattern);
	vsendto_channel_butone(one, from, chptr, pattern, vl);
	va_end(vl);
}


// sendto_channelops_butone Added 1 Sep 1996 by Cabal95.
//       Send a message to all OPs in channel chptr that
//       are directly on this server and sends the message
//       on to the next server if it has any OPs.
//
//       All servers must have this functional ability
//        or one without will send back an error message. -- Cabal95

static void vsendto_channelops_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, va_list vl)
{
	int i;
	Link *lp;
	aClient *acptr;

	for (i = 0; i < MAXCONNECTIONS; i++)
		sentalong[i] = 0;

	for (lp = chptr->members; lp; lp = lp->next) {
		acptr = lp->value.cptr;
		if (acptr->from == one || (lp->flags & CHFL_ZOMBIE) || !(lp->flags & CHFL_CHANOP))
			continue; // ...was the one I should skip or user not not a channel op
		i = acptr->from->fd;
		if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[i] = 1;
		}
		else {
			// Now check whether a message has been sent to this remote link already
			if (sentalong[i] == 0) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = 1;
			}
		}
	}
	return;
}

void sendto_channelops_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	va_list vl;
	va_start(vl, pattern);
	vsendto_channelops_butone(one, from, chptr, pattern, vl);
	va_end(vl);
	return;
}


// sendto_channelvoice_butone
// direct port of Cabal95's sendto_channelops_butone
// to allow for /notice @+#channel messages
// not exactly the most adventurous coding (made heavy use of copy-paste) <G>
// but it's needed to avoid mass-msg trigger in script vnotices
// -DuffJ

static void vsendto_channelvoice_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, va_list vl)
{
	int i;
	Link *lp;
	aClient *acptr;

	for (i = 0; i < MAXCONNECTIONS; i++)
		sentalong[i] = 0;

	for (lp = chptr->members; lp; lp = lp->next) {
		acptr = lp->value.cptr;
		if (acptr->from == one || (lp->flags & CHFL_ZOMBIE) || !((lp->flags & CHFL_VOICE) || (lp->flags & CHFL_CHANOP)))
			continue; // ...was the one I should skip or user not (a channel voice or op)
		i = acptr->from->fd;
		if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[i] = 1;
		}
		else {
			// Now check whether a message has been sent to this remote link already */
			if (sentalong[i] == 0) {
				sendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = 1;
			}
		}
	}
	return;
}

void sendto_channelvoice_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	va_list vl;
	va_start(vl, pattern);
	vsendto_channelvoice_butone(one, from, chptr, pattern, vl);
	va_end(vl);
	return;
}


// sendto_server_butone
// Send a message to all connected servers except the client 'one'.

void sendto_serv_butone(aClient *one, char *pattern, ...)
{
	int i;
	va_list vl;
	aClient *cptr;

	va_start(vl, pattern);

#ifdef NPATH
	check_command((long)2, pattern, p1, p2, p3);
#endif

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || (one && cptr == one->from))
			continue;
		if (IsServer(cptr))
			vsendto_one(cptr, pattern, vl);
	}
	va_end(vl);
	return;
}


// sendto_serv_butone_quit
//
// Send a message to all connected servers except the client 'one'.
// BUT, don't send to NOQUIT servers.

void sendto_serv_butone_quit(aClient *one, char *pattern, ...)
{
	int i;
	va_list vl;
	aClient *cptr;

	va_start(vl, pattern);

#ifdef NPATH
	check_command((long)2, pattern, p1, p2, p3);
#endif

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || (one && cptr == one->from))
			continue;
		if (IsServer(cptr) && !DontSendQuit(cptr))
			vsendto_one(cptr, pattern, vl);
	}
	va_end(vl);
	return;
}

// sendto_common_channels()
// Sends a message to all people (inclusing user) on local server who are
// in same channel with user.
void sendto_common_channels(aClient *user, char *pattern, ...)
{
	int i;
	va_list vl;
	Link *lp, *lp2;
	aClient *cptr;

	va_start(vl, pattern);
	bzero(sentalong, sizeof(sentalong));

	if (user->fd >= 0)
		sentalong[user->fd] = 1;

	for (lp = user->user->channel; lp; lp = lp->next)
		for (lp2 = lp->value.chptr->members; lp2; lp2 = lp2->next) {
			cptr = lp2->value.cptr;
			if (!MyConnect(cptr) || sentalong[cptr->fd])
				continue;
			sentalong[cptr->fd]=1;
			vsendto_prefix_one(cptr, user, pattern, vl);
		}

	// well... small mistake here, if command NICK comes from server
	// we cannot send back acknowledge for nick change.
	// Thx to codemastr -- Borg88
	if(!MyConnect(user))
		return;
	vsendto_prefix_one(user, user, pattern, vl);
	va_end(vl);
	return;
}


static void vsendto_channel_butserv(aChannel *chptr, aClient *from, char *pattern, va_list vl)
{
	int i;
	struct SLink *lp;
	aClient *acptr;

	for (lp = chptr->members; lp; lp = lp->next)
		if (MyConnect(acptr = lp->value.cptr) && !(lp->flags & CHFL_ZOMBIE))
			vsendto_prefix_one(acptr, from, pattern, vl);
}


/* Send a message to all members of a channel that are connected to this server. */
void sendto_channel_butserv(aChannel *chptr, aClient *from, char *pattern, ...)
{
	va_list vl;
	va_start(vl, pattern);
	vsendto_channel_butserv(chptr, from, pattern, vl);
	va_end(vl);
}


/* send a message to a system channel */
void sendto_syschan(const char *channel, const char *pattern, ...)
{
	va_list vl;
	char buf[512], chname[32];
	aChannel *chptr;

	snprintf(chname, sizeof(chname), "&ircd.%s", channel);
	chname[sizeof(chname) - 1] = '\0';

	chptr = find_channel(chname, (aChannel*)NULL);
	if (chptr == (aChannel*)NULL)
		return;

	va_start(vl, pattern);
	vsnprintf(buf, sizeof(buf), pattern, vl);
	va_end(vl);
	buf[sizeof(buf) - 1] = '\0';

	sendto_channel_butone(NULL, &me, chptr, ":%s NOTICE %s :%s", me.name, chname, buf);
	/* sendto_channel_butserv(chptr, &me, ":%s NOTICE %s :%s", me.name, chname, buf); */
}


// send a msg to all ppl on servers/hosts that match a specified mask
// (used for enhanced PRIVMSGs)
// addition -- Armin, 8jun90 (gruner@informatik.tu-muenchen.de)

static int match_it(aClient *one, char *mask, int what)
{
	switch (what) {
		case MATCH_HOST:
			return (match(mask, one->user->host)==0);
		case MATCH_SERVER:
		default:
			return (match(mask, one->srvptr->name)==0);
	}
}


// sendto_match_servs
//
// send to all servers which match the mask at the end of a channel name
// (if there is a mask present) or to all if no mask.
void sendto_match_servs(aChannel *chptr, aClient *from, char *format, ...)
{
	int i;
	char *mask;
	va_list vl;
	aClient *cptr;

	va_start(vl, format);

# ifdef NPATH
	check_command((long)3, format, p1, p2, p3);
# endif
	if (chptr) {
		if (*chptr->chname == '&')
			return;
		if (mask = strrchr(chptr->chname, ':'))
			mask++;
	}
	else
		mask = (char *)NULL;

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]))
			continue;
		if ((cptr == from) || !IsServer(cptr))
			continue;
		if (!BadPtr(mask) && IsServer(cptr) && match(mask, cptr->name))
			continue;
		vsendto_one(cptr, format, vl);
	}
	va_end(vl);
}

/*
 * sendto_match_butone
 *
 * Send to all clients which match the mask in a way defined on 'what';
 * either by user hostname or user servername.
 */
void sendto_match_butone(aClient *one, aClient *from, char *mask, int what, char *pattern, ...)
{
	int  i;
	va_list vl;
	aClient *cptr, *acptr;
	char cansendlocal, cansendglobal;

	va_start(vl, pattern);

	if (MyConnect(from)) {
		cansendlocal = (OPIsLocOp(from)) ? 1 : 0;
		cansendglobal = (OPIsOper(from)) ? 1 : 0;
	}
	else
		cansendlocal = cansendglobal = 1;

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]))
			continue; // that clients are not mine

		if (cptr == one) // must skip the origin !!
			continue;

		if (IsServer(cptr)) {
			if (!cansendglobal)
				continue;
			for (acptr = client; acptr; acptr = acptr->next)
				if (IsRegisteredUser(acptr) && match_it(acptr, mask, what) && acptr->from == cptr)
					break;
			// a person on that server matches the mask, so we
			// send *one* msg to that server ...
			if (acptr == NULL)
				continue;

			// ... but only if there *IS* a matching person
		}
		// my client, does he match ?
		else if (!cansendlocal || (!(IsRegisteredUser(cptr) && match_it(cptr, mask, what))))
			continue;
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

// sendto_all_butone.
// Send a message to all connections except 'one'. The basic wall type
// message generator.
void sendto_all_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	va_list vl;
	aClient *cptr;

	for (va_start(vl, pattern), i = 0; i <= highest_fd; i++)
		if ((cptr = local[i]) && !IsMe(cptr) && one != cptr)
			sendto_prefix_one(cptr, from, pattern, vl);

	va_end(vl);
	return;
}


// sendto_ops
//      Send to//local* ops only.
void sendto_ops(char *pattern, ...)
{
	int i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);

	for (i = 0; i <= highest_fd; i++)
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && SendServNotice(cptr)) {
			sprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
		}

	va_end(vl);
	return;
}

// sendto_failops
//      Send to//local* mode +g ops only.
void sendto_failops(char *pattern, ...)
{
	int i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);

	for (i = 0; i <= highest_fd; i++)
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && SendFailops(cptr)) {
		sprintf(nbuf, ":%s NOTICE %s :*** Global -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
		vsendto_one(cptr, nbuf, vl);
	}

	va_end(vl);
	return;
}

// sendto_umode
// Send to specified umode

void sendto_umode(int umodes, char *pattern, ...)
{
	int  i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);

	for (i = 0; i <= highest_fd; i++) {
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && (cptr->umodes & umodes)==umodes) {
			sprintf(nbuf, ":%s NOTICE %s :", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
		}
	}

	va_end(vl);
	return;
}


// sendto_failops_whoare_opers
// Send to//local* mode +g ops only who are also +o.

void sendto_failops_whoare_opers(char *pattern, ...)
{
	int i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);

	for (i = 0; i <= highest_fd; i++) {
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && SendFailops(cptr) && IsAnOper(cptr)) {
			sprintf(nbuf, ":%s NOTICE %s :*** Global -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
		}
	}

	va_end(vl);
	return;
}


// sendto_locfailops
// Send to//local* mode +g ops only who are also +o.
void sendto_locfailops(char *pattern, ...)
{
	int i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);

	for (i = 0; i <= highest_fd; i++) {
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && SendFailops(cptr) && IsAnOper(cptr)) {
			sprintf(nbuf, ":%s NOTICE %s :*** LocOps -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
		}
	}

	va_end(vl);
	return;
}


// sendto_opers
//         Send to//local* ops only. (all +O or +o people)

void sendto_opers(char *pattern, ...)
{
	int i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);

	for (i = 0; i <= highest_fd; i++) {
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && IsAnOper(cptr)) {
			sprintf(nbuf, ":%s NOTICE %s :*** Oper -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
		}
	}

	va_end(vl);
	return;
}


// sendto_ops_butone
//         Send message to all operators.
// one - client not to send message to
// from- client which message is from *NEVER* NULL!!

void sendto_ops_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int  i;
	va_list vl;
	aClient *cptr;

	va_start(vl, pattern);

	for (i=0; i <= highest_fd; i++)
		sentalong[i] = 0;

	for (cptr = client; cptr; cptr = cptr->next) {
		if (!SendWallops(cptr))
			continue;

		i = cptr->from->fd; // find connection oper is on

		if (sentalong[i]) // sent message along it already ?
			continue;

		if (cptr->from == one)
			continue; // ...was the one I should skip

		sentalong[i] = 1;
		vsendto_prefix_one(cptr->from, from, pattern, vl);
	}

	va_end(vl);
	return;
}

// sendto_ops_butone
//         Send message to all operators regardless of whether they are +w or
//         not..
// one - client not to send message to
// from- client which message is from *NEVER* NULL!!
void sendto_opers_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	va_list vl;
	aClient *cptr;

	va_start(vl, pattern);

	for (i=0; i <= highest_fd; i++)
		sentalong[i] = 0;

	for (cptr = client; cptr; cptr = cptr->next) {
		if (!IsAnOper(cptr))
			continue;

		i = cptr->from->fd; // find connection oper is on

		if (sentalong[i]) // sent message along it already ?
			continue;

		if (cptr->from == one)
			continue; // ...was the one I should skip

		sentalong[i] = 1;

		// This is a HORRIBLE hack.  but it's the only way I could think of to
		// format the notice messages properly.  --Russell
		vsendto_prefix_one(cptr->from, from, pattern, vl);
	}

	va_end(vl);
	return;
}

// sendto_ops_butme
// Send message to all operators except local ones
// from- client which message is from *NEVER* NULL!!
void sendto_ops_butme(aClient *from, char *pattern, ...)
{
	int i;
	va_list vl;
	aClient *cptr;

	va_start(vl, pattern);

	for (i=0; i <= highest_fd; i++)
		sentalong[i] = 0;

	for (cptr = client; cptr; cptr = cptr->next) {
		i = cptr->from->fd; // find connection oper is on

		if (!SendWallops(cptr) || sentalong[i] || cptr->srvptr == &me)
			continue;

		sentalong[i] = 1;
		vsendto_prefix_one(cptr->from, from, pattern, vl);
	}

	va_end(vl);
	return;
}

void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
{
	va_list vl;
	va_start(vl, pattern);
	vsendto_prefix_one(to, from, pattern, vl);
	va_end(vl);
}


/* send a notice to all local opers, regardless of the +s usermode */
void sendto_realops(char *pattern, ...)
{
	int i;
	va_list vl;
	char nbuf[1024];
	aClient *cptr;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr) && IsOper(cptr)) {
			sprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
			strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
			vsendto_one(cptr, nbuf, vl);
		}
	}
	va_end(vl);
	return;
}


/* send usage specification to an user */
void send_syntax(aClient *cptr, const char *command, const char *pattern, ...)
{
	char nbuf[512];
	va_list vl;

	sprintf(nbuf, ":%s 461 %s %s :Not enough parameters, usage: ", me.name, cptr->name, command);
	strcat(nbuf, pattern);

	va_start(vl, pattern);
	vsendto_one(cptr, nbuf, vl);
	va_end(vl);
}
